%This function take the feature sets f1,f2, matches them according to NN  and uses the coresponding locations to do a geometric check. Returned is a score on how well the features agree with the "best" transformation, along with some other things.
%INPUT
%f#(:,i) - is feature vector i
%l#(:,i) - is [row column scale rotation]' of feature i
%OUTPUT
%initError - the agreement on the best affine (ie. scale, rotation and shift) transformation between image feature sets. This score is [0 1] and basically means the percentage of features that agree with this transformation. How much a feature agrees with this transformation is determine by the parameters h below
%bestTrans(:) = [row col scale rotation] the best transformation to geometrically align feature set 1 to feature set 2
%errors - for diagnostics
%unnormerr - Is the same as initError, except we don't divide by the number of features. This basically measures how many features agree with the best transformation.

%Note: Given [x1,y1,s1,o1], the location, scale and orientation of feature 1, and [x2,y2,s2,o2] for feature 2, and for a given transformation feature 1 is [tx1,ty1,ts1,to1], the error the transformation is the product of three error values
%scale = exp(- ((log(ts1/s2))^2) / h(1) )
%ori = exp(- ((to1-o2)^2) / h(2) ) it handles the wrap around nature of orientation
%trans = exp(- ((tx1-x2)^2+(ty1-ty2)^2) / h(3) )
function [initialError,bestTrans,errors,unnormerr] = imageMatch(f1,f2,l1,l2)
	modelSize = [max(l1(1,:))-min(l1(1,:)) max(l1(2,:))-min(l1(2,:))] ;
	
	%[scaleSigma^2 oriSigma^2 translationSigma^2]
	%These parameters weight how much scale, orientation, and translation need to match between a feature match for a given transformation
	h = [0.16 .3 (mean(modelSize)/20)^2];
	
	%Find the best matches between the feature sets
	dists = getDist(f1,f2)';
	[v1,i1] = min(dists); %For best match, the distance whoo they match for each feature
	dists(sub2ind(size(dists),i1,1:length(i1)))=100000000;
	[v2,i2] = min(dists); %For 2nd best match, the distance whoo they match for each feature
	values = [v1;v2];
	index = [i1;i2];

	%Basically reducing the number of matches, for time consideration, by removing those that are not significantly closer than the 2nd best match, and also remove those who's difference in scale is too great (we punt if object sizes are 10 times smaller or larger)
	rescale =  abs(log(l2(3,i1)./l1(3,:)));
	good = find((values(1,:)./values(2,:))<0.8 & rescale<2);
	thesame = diff(l1(1,good));
	good = good(thesame~=0);

	%Here search the space of transformations by simply using the transformation suggested by each match. Each pair has a scale, ori difference which are determined by the features, and using these we can infer the translation. Using this transformation, we can determine how much the rest of the features agree.
	[serr,oerr,terr,subset,ex]=	findFastMatch(l1(:,good),l2(:,index(1,good)));

	tmpError = exp(-(serr/h(1)+oerr/h(2)+terr/h(3))); %Still need to remove multiple matches from feature set 1 to a feature in set 2
	theError=0;

	%Removing multiple matches in this section
	uniqueMatches = unique(index(1,good));
	if length(uniqueMatches)<1
		uniqueMatches=[];
	end
	for i=uniqueMatches
		where = find(index(1,good)==i);
		if length(where)>1
			theError = theError + max(tmpError(where,:));
		else 
			theError = theError+tmpError(where,:);
		end
	end
	theError= theError';

	%
	[initialError,which] = max(theError);
	
	 
	unnormerr=0;
	bestTrans=[];
	errors = 0;
	if theError>0	
		%A bunch of diagnostics
		matchesIndError = exp(-[serr(:,which)/h(1) oerr(:,which)/h(2) terr(:,which)/h(3)]);
		matchesError = prod(matchesIndError')';
		extra = [l1(:,good)' l2(:,index(1,good))'];
		errors = [matchesIndError matchesError extra];
		
		%This is the best trans found, or at least the best approximation
		bestTrans = ex(:,which);
		
		unnormerr= initialError;
		initialError = initialError/(size(l1,2));
		
		better = find(matchesError>0.1);		
	
	end


%each feature match L1(:,i) to L2(:,i),suggests a transformation ti, and this function determines the errors for each feature match if the transformation is applied.
%In others words, apply the transformation and determine how far the transformaed L1 is from L2
%INPUT
%l#(:,i) - is [row column scale rotation]' of feature i
%OUTPUT:
%Note: each feature match L1(:,i) to L2(:,i),suggests a transformation ti 
%rescaleError(j,i) - squared error for scale of feature j under transformation ti
%reoriError(j,i) - squared error for orientation of feature j under transformation ti
%transError(j,i) - squared error for translation of feature j under transformation ti
%ex(:,i) is the transformation suggested by match between L1(:,i) and L2(:,i)
function [rescaleError,reoriError,transError,subset,ex] = findFastMatch(L1,L2)

	ex=0;
	rescale =  L2(3,:)./L1(3,:);
	reori = mod(L2(4,:)-L1(4,:) + pi,2*pi)-pi;
	
	trans1 = rescale([1 1],:).*[cos(reori);-sin(reori)];
	trans2 = rescale([1 1],:).*[sin(reori);cos(reori)];

	shift =[sum(trans1.*L1([1 2],:));sum(trans2.*L1([1 2],:))] - L2([1 2],:);
	
	transError = (trans1'*L1([1 2],:) - repmat(shift(1,:)',1,length(reori)) -repmat(L2(1,:),length(reori),1)).^2 + (trans2'*L1([1 2],:) - repmat(shift(2,:)',1,length(reori)) - repmat(L2(2,:),length(reori),1)).^2;
		
	reoriError = repmat(reori,length(reori),1);
	reoriError =abs((mod(reoriError-reoriError'+pi,2*pi)-pi)).^2;
	rescaleError = abs(log(rescale'*(1./rescale))).^2;
	transError = transError';

	transError = transError./repmat(rescale.^2,length(rescale),1);
	
	ex = [rescale;reori;shift];
		
	subset = 1:size(L1,2);

function trans = getPerAffine(L1,L2) 
	rescale =  L2(3,:)./L1(3,:);
	reori = mod(L2(4,:)-L1(4,:) + pi,2*pi)-pi;
	trans1 = rescale([1 1],:).*[cos(reori);-sin(reori)];
	trans2 = rescale([1 1],:).*[sin(reori);cos(reori)];
	shift =[sum(trans1.*L1([1 2],:));sum(trans2.*L1([1 2],:))] - L2([1 2],:);
	trans = [rescale;reori;shift];

%This distance function is to measure the distance between features in f1 and f2. Assumes SIFT here, but any distance function can be here.
function dists = getDist(f1,f2)
	dists = 2*512*512-2*f1'*f2;