%
% Function computePanTilt takes a 3-D point in the robot's coordinate frame
% and approximates the pan and tilt values that will allow the robot to
% look at this point.
%
% David Meger - July 19, 2007
%
% isForPhoto - true if we are trying to take a picture (use best camera)
%            - false otherwise (use bumblebee)
% objLoc - Object location in coordinates relative to the robot
% position_offset - offset between robot's position and camera's positions

function [ state valid ] = computePanTilt(isForPhoto, objLoc, position_offset, cameraPars, configPars)

state = [];
if isForPhoto
    if cameraPars.canonfl
        CAMERAHEIGHT = cameraPars.CC_HEIGHT35;
    else
        CAMERAHEIGHT = cameraPars.BB_HEIGHT35;
    end;
else
    CAMERAHEIGHT = cameraPars.BB_HEIGHT35;
end;
ptu_lim = cameraPars.ptu_lim;
max_dist = cameraPars.MAX_LOOK_DIST;
tilt_offset = configPars.tilt_offset;

objLoc(1) = objLoc(1) - position_offset;

% First, compute the pan and tilt, regardless of error cases

pan_rad = atan2(objLoc(2),objLoc(1)); % PTU pan
horizDist = abs(objLoc(1)/cos(pan_rad));
tilt_rad = atan2((CAMERAHEIGHT-objLoc(3)),horizDist);

% Now several checks to ensure that the view will be valid

pan_deg = pan_rad *180/pi;
tilt_deg = -tilt_rad *180/pi + tilt_offset;

if pan_deg < ptu_lim(1) || pan_deg > ptu_lim(2) || tilt_deg < ptu_lim(3) || tilt_deg > ptu_lim(4)
    fprintf('pan/tilt %d %d exceed limits!\n',pan_deg,tilt_deg);
    valid = false;
    return;
end;

minDist = (CAMERAHEIGHT-objLoc(3))/tan(abs(ptu_lim(3)-tilt_offset)*pi/180);
maxDist = max_dist*cos(tilt_rad);

if objLoc(1) < 0.1 || horizDist < minDist || horizDist > maxDist
    
    fprintf('Attempted to look at an object too close or far.\n');
    valid = false;
    return;
end;

valid = true;
state = [ pan_deg ; tilt_deg ];
