/*
**  fullsearch_neighbours.c
**  Matlab MEX implementation of extensive search for K 
**  least squares nearest neighbours
**  Compile command: mex -DDOUBLE_FLOATS fullsearch_neighboursmex.c mexutil.c image_buffer.c
*/

#include "mex.h"
#include "mexutil.h"
#include "image_buffer.h"


void fullsearch_neighbours(buffer *bf_xp,buffer *bf_xq,int K,
			   ibuffer *bf_indl,buffer *bf_residuals) {

  buffer *bf_dist2;
  int D,N,k,l,minind,tmpi;
  int *indl,*indl2;
  fpnum *dist2,*xp,*xq,*residuals;
  fpnum diff,minval,tmpf;

  D = bf_xp->rows;
  N = bf_xp->cols;
  xp = bf_xp->data;
  xq = bf_xq->data;
  residuals = bf_residuals->data;
  indl = bf_indl->data;

  /* Generate index list */
  indl2=(int *)calloc(N,sizeof(int));
  for(k=0;k<N;k++) indl2[k]=k;

  /* Compute distances */
  bf_dist2=buffer_new(N,1,1);
  dist2 = bf_dist2->data;
  for(k=0;k<N;k++) {
    for(l=0;l<D;l++) {
      diff = xp[k*D+l]-xq[l];
      dist2[k] += diff*diff;
    }
  }
  
  /* Insert sort for K nearest neighbours */
  for(k=0;k<K;k++) {
    minval=1e20;
    minind=0;
    for(l=k;l<N;l++) {
      if(dist2[l]<minval) {
	minind=l;
	minval=dist2[l];
      }
    }
    tmpf=dist2[k];
    dist2[k]=dist2[minind];
    dist2[minind]=tmpf;
    tmpi=indl2[k];
    indl2[k]=indl2[minind];
    indl2[minind]=tmpi;
  }

  /* Copy dist and indl to output */
  for(k=0;k<K;k++) {
    residuals[k]=dist2[k];
    indl[k]=indl2[k]+1;  /* Convert to matlab indices */
  }

  buffer_free(bf_dist2);
  free(indl2);
}

/*
**  mexFunction is the interface function to MATLAB.
**
** plhs - pointer to left-hand OUTPUT mxArrays
** nlhs - number of left-hand OUTPUT arrays
** prhs - pointer to right-hand INPUT mxArrays
** nrhs - number of right-hand INPUT arrays
**
*/
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[])
{
 int msize[]={1,1};       /* Matrix size of scalar */
 int N1,K;
 buffer *bf_x,*bf_xq,*bf_residuals;
 ibuffer *bf_indl;

  /* Argument count check */
 if(nrhs==0) goto mexErrExit;
 if(nrhs != 3) {
   mexPrintf("Error: 3 input arguments expected.\n");
   goto mexErrExit;
 }

  /* Parse arg 1 */
 msize[0]=-1;msize[1]=-1;
  if(!type_check("X",prhs[0],mxDOUBLE_CLASS,2,msize))
    goto mexErrExit;
  N1=mxGetM(prhs[0]);
  bf_x=buffer_encapsulate(prhs[0]);

  /* Parse arg 2 */
 msize[0]=N1;msize[1]=-1;
  if(!type_check("xq",prhs[1],mxDOUBLE_CLASS,2,msize))
    goto mexErrExit;
  bf_xq=buffer_encapsulate(prhs[1]);

  /* Parse arg 3 */
 msize[0]=1;msize[1]=1;
  if(!type_check("K",prhs[2],mxDOUBLE_CLASS,2,msize))
    goto mexErrExit;
  K=mxGetScalar(prhs[2]);

  plhs[0]=mxCreateNumericMatrix(K,1,mxINT32_CLASS,mxREAL);
  bf_indl=ibuffer_encapsulate(plhs[0]);

  if(nlhs>1) {
    plhs[1]=mxCreateNumericMatrix(K,1,mxDOUBLE_CLASS,mxREAL);
    bf_residuals=buffer_encapsulate(plhs[1]);
  } else {
    bf_residuals=buffer_new(K,1,1);
  }

  /* Call computation function */
  fullsearch_neighbours(bf_x,bf_xq,K,bf_indl,bf_residuals);

  /* Free used arrays */
  if(nlhs>1) free(bf_residuals);   /* Only release struct */
  else buffer_free(bf_residuals);
  
  free(bf_indl);   /* Only release struct */
  free(bf_xq);     /* Only release struct */
  free(bf_x);      /* Only release struct */
  
/* End of show :-) */

  goto mexExit;

/* Exit with error message */
mexErrExit:
 mexPrintf("Usage: [<cl>,<r>]=fullsearch_neighbours(X,xq,K)\n");
mexExit: {}
}
