function [configPars mappingPars] = exploreLocation(cameraPars,gazePars,mappingPars,configPars,planningPars)
% Explore this location (look for objects):
% (a) grab bumblebee image and pre-process 
% (b) find K interesting regions
% (c) for region =  1:K
%
% cameraPars - camera parameters
% gazePars - gaze parameters
% mappingPars - mapping parameters
% configPars - configuration parameters
% ptu_lim - PTU settings limits 
%

% Unpack parameters
ptu_lim = cameraPars.ptu_lim;
configFile = cameraPars.configFile;
canonfl = cameraPars.canonfl;
bbtocan_scaling = cameraPars.bbtocan_scaling;
canon_res = cameraPars.canon_res;
zoom_scaling = cameraPars.zoom_scaling;
is16BitDisp = cameraPars.is16BitDisp;
res = cameraPars.res;
xx = cameraPars.xx;
yy = cameraPars.yy;
msz = cameraPars.msz;
memory = gazePars.memory;
pars = gazePars.pars;
curObjId = mappingPars.curObjId;
objs = mappingPars.objs;
usingLaptop = configPars.usingLaptop;
tilt_offset = configPars.tilt_offset;
nTotalPics = configPars.nTotalPics;
objectFile1 = configPars.objectFile1;
objectFile2 = configPars.objectFile2;
DISPLAY = configPars.DISPLAY;
output_path = configPars.output_path;
output_path2 = configPars.output_path2;

% Find views which will cause us to cover some new ground with the camera's
% FOV
doSlam(1,true,0);
[binScores wedgeMasks] = getUnseenViews( mappingPars, planningPars );
% invert sign for binCentres because y-axis in grid coordinates points down
binCentres = -[ -pi+pi/6:pi/3:pi-pi/6 ];

if nnz(binScores)>3
    
    % Pick the 3 best angles to explore
    
    [sortedScores idx] = sort(-binScores);
    binCentres = binCentres(idx(1:3));
    wedgeMasks = wedgeMasks(:,:,idx(1:3));
else
    binCentres = binCentres(find(binScores>0));
    wedgeMasks = wedgeMasks(:,:,find(binScores>0));
end;

% Display the views we want to look at
if DISPLAY
    positionPars = updatePositionPars();
    figure(10);
    colormap(gray(256));
    viewsPlot = positionPars.oGrid;
    
    wedgeMask = zeros(size(wedgeMasks(:,:,1)));
    nAngles = length(binCentres);
    for k=1:nAngles
        wedgeMask = wedgeMask | wedgeMasks(:,:,k);
    end;
    viewsPlot(wedgeMask) = max(0,viewsPlot(wedgeMask)-25);
    image(viewsPlot);
    hold on;
    plot(positionPars.rGridPos(2),positionPars.rGridPos(1),'g+');
    mappingPars.mapId = saveMap(mappingPars.mapId);
end;

while ~isempty(binCentres)
    
    % Sort binCentres by proximity to current angle and turn the robot to
    % explore that bin
    
    positionPars = updatePositionPars();
    yaw = positionPars.rPose(3);
    
    diff = [];
    
    for k=1:length(binCentres)
        diff(k) = thetapp(binCentres(k)-yaw);
    end;
    
    [sortedDiff idx] = sort(abs(diff));
    
    minRange = binCentres(idx(1)) - pi/6;
    maxRange = minRange + pi/3;
    
    binCentres = binCentres([1:idx(1)-1 idx(1)+1:length(binCentres)]);
    turnbot(diff(idx(1))*180/pi);
    doSlam(1,false);
    
    
    fprintf('Robot will look at regions between %g and %g.\n',minRange*180/pi,maxRange*180/pi);


    % Look down
    tilt0=-35;40; % Look 40deg down
    state0=powerbot('setpos',0,tilt0);
    tilt0=state0(2);

    [img,disp,objectImage obstacles] = getImages(tilt0,false,configPars,cameraPars);         % (1)
    rPose = performSlam('get_curvecpos');
    [ mappingPars.coverageMap ] = updateCoverage( mappingPars.coverageMap, -tilt0,rPose );

    % Don't look too far down
    angle_left=tilt0-ptu_lim(3);
    objectImage = pruneObjectImage(objectImage,angle_left);

    if DISPLAY,

        figure(99);
        clf;
        subplot(2,1,1);
        imagesc(mappingPars.coverageMap);
        subplot(2,1,2);
        imagesc(performSlam('get_map'));
        drawnow;

    end;

    %[ objects ] = getObjects( img, objectImage,configPars.ss.Kmax );     % (2)
    [ objects saliencyMap ] = getObjects2( img, objectImage,configPars.ss );

    if DISPLAY

        figure(1);clf;
        objectImage=double(objectImage);
        objectImage = objectImage/max(objectImage(:))*255;
        if is16BitDisp
            disp16=disp;disp16(disp16>64257)=0;disp16=8/256*disp16;
            image([ rgb2gray(img) disp16; objectImage uint16(saliencyMap*255)]);
        else
            disp8=disp;disp8(disp8>251)=0;disp8=8*disp8;
            image([ rgb2gray(img) disp8; objectImage uint8(saliencyMap*255)]);
        end;
        colormap(gray(256));
        drawnow;

        figure(2);clf;
        image(img);
        hold on;

        for obj = 1:length(objects)

            %plot(objects(obj).centre(2), objects(obj).centre(1),'r+');
            h=plot_objnum(objects);
            set(h,'Color',[1 0 0]);
            bbox = objects(obj).bb;
            h=plot_bbox(bbox,'r-');
        end;

        hold off;
    end;

    tilt = tilt0-tilt_offset;
    tiltRad = tilt * pi/180;

    % Matrix to rotate 3D coordinates from camera to robot frame
    Rx_cam = [1    0              0;
        0    cos(tiltRad)   -sin(tiltRad);
        0    sin(tiltRad)   cos(tiltRad)];

    % Add salient objects to map
    [objs nxtObjId objLocs] = addObjectsToMap(Rx_cam,disp,objects,cameraPars,mappingPars);

    [states zooms sc_tots] = getPanTiltZoom(disp,objects,objLocs,cameraPars,gazePars,configPars);

    % Now look at objects
    views=getViews(states,configFile,usingLaptop,canonfl,zooms);

    % Save acquired views to disk
    for m=1:length(objects)

        if ~isempty(views{m})

            objId = curObjId+m-1;
            % increment bin of angle that we're about to see
            rPose = performSlam('get_curvecpos');
            pan_rad = states{m}(1) *pi/180;
            angle = (rPose(3) + pan_rad) *18/pi;
            index = 5 + floor(mod(angle,36));
            objs(objId,index) = objs(objId,index) + 1;
            % Increment nPicsTaken of this
            objs(objId,4) = objs(objId,4) + 1;
            nTotalPics = nTotalPics + 1;


            fname = sprintf('%s%d.jpg',output_path,nTotalPics);
            if canonfl
                system(sprintf('mv lastframe%d.jpg %s',m,fname));
            else
                imwrite( views{m}, fname, 'JPG','Quality',95 );
            end;
            wco=[size(views{m},1); size(views{m},2)]/2; % Image centre
            bbox=centreBoundingBox(objects(m),sc_tots(m),wco);
            imb=cropImage(views{m},bbox,2);
            fname = sprintf('%s%d.jpg',output_path2,nTotalPics);
            imwrite( imb, fname, 'JPG','Quality',95 );

            % Compute SIFT in the background
            %fprintf('Spawning SIFT computation process...\n');
            %fname2 = sprintf('%s%d.sift',output_path2,nTotalPics);
            %system(sprintf('%skeypoints %s %s %d &',keypointPath,fname,fname2,3000));

            % Write id of object into objects.txt
            entry = sprintf('%d\n',objId);
            fwrite(objectFile1,entry,'char');
            fwrite(objectFile2,entry,'char');
        end;
    end;

    mappingPars.objs = objs;
    mappingPars.curObjId = nxtObjId;
    configPars.nTotalPics = nTotalPics;

    if DISPLAY
        figure(3);clf
        for m=1:min(4,length(objects)),
            subplot(2,2,m)
            if(length(views{m}))
                image(views{m});axis image
                hold on
                h=plot(xx*msz+wco(2),yy*msz+wco(1),'r');
                set(h,'LineWidth',3.0);
                bbox=centreBoundingBox(objects(m),sc_tots(m),wco);
                plot_bbox(bbox,'r-');
            else
                h=text(0,0,'Gaze failed');
                set(h,'FontSize',15);
                set(h,'HorizontalAlignment','center');
                set(h,'VerticalAlignment','middle');
                axis([-1 1 -1 1]);axis off
            end
            title(['View ' num2str(m)]);
        end
        drawnow;
    end;
end;

