%
% inputImage   RGB input image for saliency computation
% objectMask   Binary mask of valid locations for interest points
% K            Number of points to select
% params       Optional parameter for saliency computation
%
function [ objects ] = getObjects( inputImage, objectMask, k, varargin )

declareGlobal;
%DEBUG_FID=1;   % This enables verbose output from the Saliency toolbox
objects = [];
objIndex = 0;

% initialize the Image structure if necessary
if (isa(inputImage,'struct'))
  img = inputImage;
else
  img = initializeImage(inputImage);
end

% use the default saliency parameters if the user didn't specify
if isempty(varargin)
  params = defaultSaliencyParams(img.size,'dyadic');
else
  params = varargin{1};
end

% make sure that we don't use color features if we don't have a color image
if (img.dims == 2)
  params = removeColorFeatures(params);
end

% create the saliency map
[salmap,salData] = makeSaliencyMap(img,params);

% display the conspicuity maps and the saliency map
%figure; displayMaps({salData.CM salmap},2);

% initialize the winner-take-all network
wta = initializeWTA(salmap,params);

% display the input image
%imgFig = showImage(img);

% display the current WTA
%salFig = showImage(imresize(wta.sm.V,img.size(1:2),'bilinear'),...
 %                  'Winner Take All',1);

shapeFig = -1;
lastWinner = [-1,-1];
reply = '';

% loop over the successive fixations, until k are obtained,
% or we have tried 20 times
iterations = 0;
while objIndex < k & iterations < 20
  iterations = iterations+1;

  winner = [-1,-1];

  % evolve WTA until we have a winner
  while (winner(1) == -1)
    [wta,winner] = evolveWTA(wta);
  end

  % update the WTA plot
 % figure(salFig);
 % displayMap(imresize(wta.sm.V,img.size(1:2),'bilinear'),1);
  
  % run the shape estimator to get proto-objects
  shapeData = estimateShape(salmap,salData,winner,params);
  
 %figure();image(shapeData.shapeMap.data.*double(objectMask).*255);
 % figure();image(shapeData.shapeMap.data.*255);
 
  % trigger inhibition of return
  wta = applyIOR(wta,winner,params,shapeData);
  
  % Reject object if it doesn't touch the objectMask
  if( max(max(shapeData.shapeMap.data.*double(objectMask))) == 0 )
      continue;
  end;
  
  % convert the winner's location to image coordinates
  win2 = winnerToImgCoords(winner,params);
 
  % plot the currently attended region into the image figure
  %plotSalientLocation(imgFig,win2,lastWinner,params,shapeData);
  
  objIndex = objIndex + 1;
  objects(objIndex).centre = win2;
  objects(objIndex).shapeData = shapeData;
  
  [objRows,objCols]=find(shapeData.shapeMap.data>0);

  objects(objIndex).bb.topLeft = [ min(objRows) min(objCols) ];
  objects(objIndex).bb.botRight = [ max(objRows) max(objCols) ];  
  
  lastWinner = win2;
  
  % in case we have shape data, create a plot 
  % to display various processing steps
  if ~isempty(shapeData)
    
    % need to open a new figure or use the previous one?
%     if (shapeFig == -1) 
%       shapeFig = figure; 
%     else
%       figure(shapeFig);
%     end
    
    % draw the various maps
    %displayMaps({shapeData.winningMap,shapeData.segmentedMap,...
    %             shapeData.binaryMap,shapeData.shapeMap});
    
    % find the right label for the figure window
    winLabel = [' - ' shapeData.winningMap.label];
    if any(isnan(img.filename))
      lab = winLabel;
    else
      lab = [img.filename winLabel];
    end
    %set(imgFig,'Name',lab);
  else
    winLabel = '';
  end
  
  % make sure everything gets drawn
  %drawnow;
  
%   % write out the details for our winner
%   txt = sprintf('winner: %i,%i; t = %4.1f ms%s',...
%                 win2(2),win2(1),wta.exc.time*1000,winLabel);
%   
%   % wait for user input - return for next fixation, 'q' to terminate
%   reply = input(txt,'s');
end
