%function [ objects sMap ] = getObjects2( inputImage, objectMask, params )
%
% Proto-objects from spectral saliency
%
% inputImage   RGB input image for saliency computation
% objectMask   Binary mask of valid locations for interest points
% params       Parameter struct with fields:
%        Kmax         Max number of points to select
%        scale        Scale in octave pyramid to use
%        overlap      Remove small objects with more than this overlap
%        depthoverlap Require this overlap with objectMask
%        minthr       Require min within region to be above mean of
%                     intensity times this.
%
% objects      A struct with fields:
%        centre      (y,x)
%        shapeData.shapeMap.data (object map)
%        bb.topLeft  (ymin xmin)
%        bb.botRight (ymax xmax)  
%
%Per-Erik Forssen, July 2007
function [ objects sMap ] = getObjects2( inputImage, objectMask, p )

% MSER parameters
mp.min_margin=10;
mp.max_area=0.25;
mp.min_size=10;30;
mp.export_image=0; % Output greyscale images or not.

% Build image pyramid
imp=image_pyramid2(inputImage,p.scale);

% Projections to try
plist=zeros(3,3); Np=size(plist,1);
plist(1,:)=[1 1 1]/3;  % grey
plist(2,:)=[1 -1 0];   % r-g
plist(3,:)=[1 1 -2]/2; % y-b

% Filters for spectral saliency
favg=fspecial('average', 3); % Fourier domain smoothing
savg=fspecial('disk', 3);   % Spatial smoothing

% Compute spectral saliency
[Iml,Sml]=spectral_saliency(imp,plist,p.scale,favg,savg);
% Sum colour bands to a single map
Sms=(Sml{1,1}+Sml{1,2}+Sml{1,3})/3;

% Run MSER to find salient objects
img=uint8(Sms*255);

% Upsample saliency map
sMap=Sms;
for l=1:p.scale-1,
    sMap = up2(sMap);
end
  
if 0
    imean=mean(img(:));
else
    objSum = sum(objectMask(:));
    if objSum > 0
        % Only average inside objectMask
        imean=mean(mean(sMap.*objectMask))/3/objSum;
    else
        % No depth means no salient objects
        imean=max(img(:));
    end;
end;
[mvec,pvec,masks]=mser_objects2(img,mp);
  
mvec=mvec_upsample(mvec,p.scale);
keepl=find(pvec(3,:)>p.minthr*imean);   % Threshold on min(region) (better)
%keepl=find(pvec(1,:)>p.minthr*imean);  % Threshold on mean(region)
bbox=bbox_from_conic(mvec(:,keepl));
keepl2=suppress_duplicates(bbox,p.overlap); % BB overlap (Better)
%keepl2=suppress_duplicates2(masks(keepl),p.overlap);% Mask overlap

% Keep only needed objects
keepl=keepl(keepl2);
masks=masks(keepl);
mvec=mvec(:,keepl);
pvec=pvec(:,keepl);
bbox=bbox(:,keepl2);

Nobj=length(masks);  % Count proto-objects

% Upsample masks
for k=1:Nobj
  for l=1:p.scale-1,
    masks{k}=up2(masks{k});
  end
end

% Accept only objects with a certain amount of objectMask overlap
keepl=zeros(Nobj,1);
objectMask=double(objectMask);
objectMask=objectMask/max(objectMask(:)); % Ensure 0/1 values
for k=1:Nobj,
  if(sum(sum(masks{k}.*objectMask))/sum(sum(masks{k}))> p.depthoverlap )
    keepl(k)=1;
  end;
end  
keepl=find(keepl);

Nobj=length(keepl);  % Count proto-objects

% Keep only needed objects
masks=masks(keepl);
mvec=mvec(:,keepl);
pvec=pvec(:,keepl);
bbox=bbox(:,keepl);

[dummy,indl]=sort(pvec(2,:)); % Sort according to max value
masks=masks(indl);
mvec=mvec(:,indl);
pvec=pvec(:,indl);
bbox=bbox(:,indl);

% Now, build the objects struct
Nobj=min(p.Kmax,Nobj);
objects = [];

for k=1:Nobj,
  objects(k).centre = [mvec(3,k) mvec(2,k)];      % Ellipse centre
  objects(k).shapeData.shapeMap.data=masks{k};    % Object mask
  objects(k).bb.topLeft  = [bbox(1,k) bbox(2,k)]; % ymin xmin
  objects(k).bb.botRight = [bbox(3,k) bbox(4,k)]; % ymax xmax
end

