function map = performMapping(tilt,disp,objects,configFile,is16BitDisp,logFile)
% Function for obtaining 2D global coordinates of odometry,
% laser readings, and salient objects
%
% tilt - PTU tilt from which the disparity was obtained
% disp - 8bit or 16bit disparity map
% objects - salient objects
% configFile - The bumblebee config file
% is16BitDisp - 1 if disparity is 16bit, 0 if it's 8bit
% logFile - The map log file
%

% Grab odometry and laser data
odom = powerbot('getodom');
laser = powerbot('getlaser');

crStr = sprintf('\n');
%log odometry
odomStr = sprintf('o %g %g %g\n',odom);
fwrite(logFile,odomStr,'char');

%log laser
fwrite(logFile,'l ','char');
if length(laser) > 0
  laserStr = sprintf('%g ',laser);
  fwrite(logFile,laserStr,'char');
end;
fwrite(logFile,crStr,'char');

% Initialize map
map.odom = odom;
map.laserPts = [];
map.objPts = [];

yaw = odom(3); % yaw in radians

%Matrix to rotate 2d coordinates from robot to global frame
R_r = [cos(yaw), -sin(yaw); sin(yaw), cos(yaw)];

% Transform laser readings from robot to global frame
count = length(laser);
for i=1:2:count
  xy_r = [laser(i), laser(i+1)]';
  xy = R_r*xy_r;
  xy = xy + [odom(1) odom(2)]';

  map.laserPts = [map.laserPts; xy'];
end; % for

% Plot salient objects onto map
tilt = tilt-5; % Add offset of -5, determined from PTU calibration
tiltRad = tilt * pi/180;

% Matrix to rotate 3D coordinates from camera to robot frame
Rx_cam = [1    0              0; 
          0    cos(tiltRad)   -sin(tiltRad);
          0    sin(tiltRad)   cos(tiltRad)];

% set disparity limits based on bitdepth
if is16BitDisp
  DISP_MAX = 65280;
  DISP_MIN = 1536;
else
  DISP_MAX = 252;
  DISP_MIN = 6;
end;

objNum = 0;
for obj=objects
  objNum = objNum+1;

  %topRight = [bbox.topLeft(1) bbox.botRight(2)];
  %botLeft = [bbox.botRight(1) bbox.topLeft(2)];

  % Calculate center of bounding box
  cen_x = (obj.bb.botRight(1)+obj.bb.topLeft(1))/2;
  cen_y = (obj.bb.botRight(2)+obj.bb.topLeft(2))/2;
  
  % Find the median disparity in the neighbourhood of the center
  disparities = [];
  for x=-5:5
    for y=-5:5
      disparity = disp(round(cen_x+x),round(cen_y+y));
      if disparity < DISP_MAX & disparity > DISP_MIN
        disparities = [disparities; disparity];
      end;
    end;
  end;
  disp(round(cen_x), round(cen_y)) = median(disparities);

  % Get object center in camera coordinates
  if is16BitDisp
    xyz_c = triclops_getXYZ16([round(cen_x) round(cen_y)], disp, configFile);
  else
    xyz_c = triclops_getXYZ([round(cen_x) round(cen_y)], disp, configFile);
  end;

  if(xyz_c(3) == -1)
    fprintf('obj=%d disp=%d Failed to get xyz_c\n',objNum,disp(round(cen_x), round(cen_y)));
  else
    % log object center
    fwrite(logFile,'s ','char');
    objStr = sprintf('%g ',xyz_c);
    fwrite(logFile,objStr,'char');
    fwrite(logFile,crStr,'char');

    % Translate object center to global coordinates
    yzx_r = Rx_cam*xyz_c';
    xy = R_r*[yzx_r(3) -yzx_r(1)]';
    xy = xy + [odom(1) odom(2)]';

    map.objPts = [map.objPts; xy'];
  end;
end;
