/*
 * laserobstacleavoid.cc
 *
 * a simple obstacle avoidance demo
 *
 * @todo: this has been ported to libplayerc++, but not tested
 */

#include <libplayerc++/playerc++.h>
#include <iostream>

#include "lodo.h"
#include "pmap.h"
#include "rmap.h"
#include "omap.h"

extern "C"{
  /*#include <time.h> */
#include <sys/time.h>
}


#define DEBUG


#include "args.h"

#define RAYS 32

// Number of seconds to go before stopping
#define RUN_TIME 5.0

double accurate_clock() {
  struct timeval tp;
  gettimeofday(&tp,NULL);
  return tp.tv_sec + tp.tv_usec/1e6;
}

// SLAM constants
static const int NUMRANGES = 361;
static const double RANGEMAX = 8.0;
static const double RANGERES = 0.1;
static const double RANGESTART = -M_PI/2;
static const double RANGESTEP = M_PI/360;
static const int SAMPLESLEN = 1000;
static const double GRIDWIDTH = 32.0;
static const double GRIDHEIGHT = 24.0;
static const double GRIDSCALE = 0.1;

// SLAM globals
lodo_t *lodo;
pmap_t *pmap;
rmap_t *rmap;
omap_t *omap;

using namespace PlayerCc;

void initializeSlam(Position2dProxy *pp, LaserProxy *lp)
{
  lodo = lodo_alloc(NUMRANGES, RANGEMAX, RANGERES,
                   RANGESTART, RANGESTEP);

  lodo->laser_pose.pos.x = 0.1;
  lodo->laser_pose.pos.y = 0;
  lodo->laser_pose.rot = 0;

  pmap = pmap_alloc(NUMRANGES, RANGEMAX,
                   RANGESTART, RANGESTEP, SAMPLESLEN,
                   GRIDWIDTH, GRIDHEIGHT, GRIDSCALE);

  rmap = rmap_alloc(NUMRANGES, RANGEMAX,
                   RANGESTART, RANGESTEP,
                   GRIDWIDTH, GRIDHEIGHT);

  omap = omap_alloc(NUMRANGES, RANGEMAX,
                   RANGESTART, RANGESTEP,
                   GRIDWIDTH, GRIDHEIGHT, GRIDSCALE);

  // Get Initial Odometry
  pose2_t pose;
  pose.pos.x = pp->GetXPos();
  pose.pos.y = pp->GetYPos();
  pose.rot = pp->GetYaw();

  pmap_set_pose(pmap, pose);
}

void stopSlam()
{
  lodo_free(lodo);
  pmap_free(pmap);
  rmap_free(rmap);
  omap_free(omap);
}

void doSlam(Position2dProxy *pp, LaserProxy *lp)
{
  // Get laser readings
  uint count = lp->GetCount();
  if(count != 361)
    return;

  double ranges[NUMRANGES];
  
  int j=0;
  for(uint i=0; i < count; i+=1)
  {
    ranges[j] = lp->GetRange(i);
    ++j;
  }

  // Get Odometry
  pose2_t pose;
  pose.pos.x = pp->GetXPos();
  pose.pos.y = pp->GetYPos();
  pose.rot = pp->GetYaw();

  // Get laser-stabilized odometry
  pose2_t lodoPose = lodo_add_scan(lodo, pose, NUMRANGES, ranges);
  
  // Update pmap
  lodoPose = pose2_add(lodo->laser_pose, lodoPose);
  pmap_update(pmap, lodoPose, NUMRANGES, ranges);
}

// Save the coarse map
void save_pmap()
{
  int i;
  char filename[256];
  pmap_sample_t *sample;
  pmap_scan_t *scan;
  pose2_t pose;

  sample = PMAP_GET_SAMPLE(pmap, pmap->best_sample);

  omap_clear(omap);

  // Create map and write trajectory
  for (i = 0; i < pmap->step_count; i++)
  {
    scan = pmap->scans + i;
    pose = sample->poses[i];

    omap_add(omap, pose, pmap->num_ranges, scan->ranges);
  }

  // Write map
  snprintf(filename, sizeof(filename), "coarse.pgm");
  fprintf(stderr, "saving %s\n", filename);  
  omap_save_pgm(omap, filename);
  
  return;
}

// Fine map generation (relax in-memory map)
void build_rmap()
{
  int i;
  pose2_t pose;
  pmap_scan_t *scan;
  pmap_sample_t *sample;

  rmap_clear(rmap);

  // Get the current best sample
  sample = PMAP_GET_SAMPLE(pmap, pmap->best_sample);

  // If the map is empty, initialize it
  for (i = 0; i < pmap->step_count; i++)
  {
    scan = pmap->scans + i;
    pose = sample->poses[i];
    rmap_add(rmap, pose, pmap->num_ranges, scan->ranges);
  }

  // Match points periodically; HACK
  //if (fine_count % 10 == 0)
  rmap_match(rmap);

  // Relax map
  rmap_relax(rmap, 10);

  // Interpolate
  rmap_interpolate(rmap);

  // Show progress
  fprintf(stderr, "%d %d %d %f \r",
          rmap->num_scans, rmap->num_key_scans, rmap->num_cons, rmap->relax_err);
  fflush(stderr);
}


// Save the fine map
void save_rmap()
{
  int i;
  char filename[256];
  rmap_scan_t *scan;
  pose2_t pose;
 
  omap_clear(omap);
    
  // Create map and write trajectory
  for (i = 0; i < rmap->num_scans; i++)
  {
    scan = rmap->scans + i;
    pose = scan->pose;
    omap_add(omap, pose, rmap->num_ranges, scan->ranges);
  }

  // Write map
  snprintf(filename, sizeof(filename), "fine.pgm");
  fprintf(stderr, "saving %s\n", filename);
  omap_save_pgm(omap, filename);
  
  return;
}

int
main(int argc, char **argv)
{
  parse_args(argc,argv);
  //time_t startTime, currentTime;
  double startTime, currentTime;
  double newspeed = 0;
  double newturnrate = 0;
  player_pose_t robotPose;
  player_pose_t robotVel;

  //char c;
  std::string garbage;

  // we throw exceptions on creation if we fail
  try
  {
    PlayerClient robot(gHostname, gPort);
    Position2dProxy pp(&robot, gIndex);
    LaserProxy lp(&robot, gIndex);

    robot.SetDataMode(PLAYER_DATAMODE_PULL);
    robot.SetReplaceRule(true, -1, -1, -1);

    std::cout << robot << std::endl;

    pp.SetMotorEnable (true);

    newturnrate = 20;
    newturnrate = dtor(newturnrate);

    pp.ResetOdometry();

    robotPose.px = 0.0;
    robotPose.py = 0.0;
    robotPose.pa = dtor(90.0);

    robotVel.px = 0.0;
    robotVel.py = 0.0;
    robotVel.pa = dtor(20.0);

    printf("Before goto.\n");
    pp.GoTo(robotPose,robotVel);
    printf("After goto.\n");

    // this blocks until new data comes; 10Hz by default
    robot.Read();

    // Initialize SLAM
    initializeSlam(&pp,&lp);

    int fineIter = 1;
    for(;;)
    {
	    
        // go into read-think-act loop
//         while( currentTime++ < 100  )
      //time(&startTime);  // Count seconds from here
      startTime=accurate_clock();
      currentTime=startTime;

      int coarseIter = 1;
      for(;currentTime-startTime<RUN_TIME;currentTime=accurate_clock())
        {
            double minR = 1e9;
            double minL = 1e9;
	    
            // this blocks until new data comes; 10Hz by default
            robot.Read();
            doSlam(&pp,&lp);
            if(coarseIter % 20 == 0)
            {
              // save map
              save_pmap();
            }
            ++coarseIter;

            // laser avoid (stolen from esben's java example)
            uint count = lp.GetCount();
            
        #ifdef DEBUG
//                 std:: cout << "Count: " << count << std::endl;
//             
//             for( uint j=0; j < count; j++ )
//             {
//                 std::cout << lp[j] << " at angle " << j << std::endl;
//             }

            currentTime=accurate_clock();
	    
	    //std::cout << "Current time: " << currentTime << " Time left: " << RUN_TIME-currentTime+startTime <<std::endl;
            
        #endif
       
            for (uint j=0; j < count/2; ++j)
            {
                if (minR > lp[j])
                minR = lp[j];
            }
            for (uint j = count/2; j < count; ++j)
            {
                if (minL > lp[j])
                minL = lp[j];
            }
//             std::cout << "minR: " << minR
//                         << "minL: " << minL
//                         << std::endl;
        
            double l = (1e5*minR)/500-100;
            double r = (1e5*minL)/500-100;
        
            if (l > 100)
                l = 100;
            if (r > 100)
                r = 100;
        
            newspeed = (r+l)/1e3;
        
            newturnrate = (r-l);
            newturnrate = limit(newturnrate, -40.0, 40.0);
            newturnrate = dtor(newturnrate);
        
            //std::cout << "speed: " << newspeed
            //            << "turn: " << newturnrate
            //            << std::endl;
        
            if( newspeed < 0.1 && newturnrate < 0.1 )
            {
                // Stalled,  turn 180
                for( int i=0; i<10; i++ )
                {
                    newturnrate = 20;
                    pp.SetSpeed(0, newturnrate);
                }
            }
        
            // write commands to robot
            pp.SetSpeed(newspeed, newturnrate);
        } // end while

        //pp.SetSpeed(0, 0);


        // this blocks until new data comes; 10Hz by default
        robot.Read();
        if(fineIter % 5 == 0)
        {
          build_rmap();
          save_rmap();
        }
        ++fineIter;
   
        //std::cout << "Waiting for input." << std::endl;
        //std::getline( std::cin, garbage);

    } // end for
  }
  catch (PlayerCc::PlayerError e)
  {
    std::cerr << e << std::endl;
    return -1;
  }
}
