/*
 *  Simple access to powerbot motion
*/

#include "pb_mini.h"
#include <libplayerc++/playerc++.h>
#include <unistd.h>
#include <iostream>

extern "C"{
#include <time.h>
}

#define DEBUG

#define RAYS 32

// Number of seconds to go before stopping
#define RUN_TIME 10.0

std::string  gHostname(PlayerCc::PLAYER_HOSTNAME);
uint         gPort(PlayerCc::PLAYER_PORTNUM);
uint         gIndex(0);
uint         gDebug(0);
uint         gFrequency(10); // Hz
uint         gDataMode(PLAYER_DATAMODE_PUSH);
bool         gUseLaser(true);

PlayerCc::PlayerClient *robot = NULL;
PlayerCc::Position2dProxy *pp = NULL;
PlayerCc::LaserProxy *lp = NULL;

int PB_init()
{
    try
    {
        //using namespace PlayerCc;

        /* In this setup, we're running the client code on the robot. Connect to loopback.*/
        gHostname = "127.0.0.1";
        
        robot = new PlayerCc::PlayerClient(gHostname, gPort);
        robot->SetDataMode(PLAYER_DATAMODE_PULL);
        robot->SetReplaceRule(true, -1, -1, -1);
        pp = new PlayerCc::Position2dProxy(robot, gIndex);
        lp = new PlayerCc::LaserProxy(robot, gIndex);
        
        std::cout << *robot << std::endl;
        pp->ResetOdometry();
        pp->SetMotorEnable(true);
    }
    catch(PlayerCc::PlayerError e)
    {
        std::cerr << e << std::endl;
        return -1;
    }

    return 1;
}

int PB_turn(float yawAmount)
{
    printf( "In PB_turn.\n" );
    // this blocks until new data comes; 10Hz by default
    robot->Read();
    yawAmount = PlayerCc::normalize(PlayerCc::dtor(yawAmount));
    double newturnrate = PlayerCc::dtor(40);
    if(yawAmount < 0)
      newturnrate = -newturnrate;

    //for( int i=0; i<10; i++ )
    // Normalize yaws to [-pi pi] so that they can be compared
    float yaw = pp->GetYaw();
    float goalYaw = yaw + yawAmount;
    float yawRemain = fabs(PlayerCc::normalize(goalYaw-yaw));
    bool isNearGoal = false;
    while(yawRemain > 0.01)
    {
        robot->Read();
        yaw = pp->GetYaw();
        yawRemain = fabs(PlayerCc::normalize(goalYaw-yaw));
        if(yawRemain < 0.4 && !isNearGoal)
        {
          newturnrate /= 4;
          isNearGoal = true;
        }
        pp->SetSpeed(0, newturnrate);
    }

    pp->SetSpeed(0, 0);

    return 1;
}

void *PB_motion(void *a)
{
    static const int NUM_ARGS = 2;
    float speedLims[NUM_ARGS];
    memcpy(speedLims,a,NUM_ARGS*sizeof(float));
    double maxSpeed = speedLims[0];
    double maxTurnRate = speedLims[1];

    time_t startTime, currentTime;

    time(&startTime);  // Count seconds from here
    
    while(difftime(time(&currentTime),startTime)<RUN_TIME)
    {
        double newspeed = 0;
        double newturnrate = 0;
        double minR = 1e9;
        double minL = 1e9;
        
        // this blocks until new data comes; 10Hz by default
        robot->Read();
        
        // laser avoid (stolen from esben's java example)
        uint count = lp->GetCount();
            
#ifdef DEBUG
//      std:: cout << "Count: " << count << std::endl;
//       
//      for( uint j=0; j < count; j++ )
//      {
//          std::cout << lp[j] << " at angle " << j << std::endl;
//      }

        time(&currentTime);
            
        std::cout << "Current time: " << currentTime << " Time left: " << RUN_TIME-difftime(currentTime,startTime) <<std::endl;
            
#endif
    
        for (uint j=0; j < count/2; ++j)
        {
            if (minR > (*lp)[j])
            minR = (*lp)[j];
        }
        for (uint j = count/2; j < count; ++j)
        {
            if (minL > (*lp)[j])
            minL = (*lp)[j];
        }

//      std::cout << "minR: " << minR
//                << "minL: " << minL
//                << std::endl;
        
        double l = (1e5*minR)/500-100;
        double r = (1e5*minL)/500-100;
    
        if (l > 200)
            l = 200;
        if (r > 200)
            r = 200;
    
        newspeed = (r+l)/1e3;
        newspeed = PlayerCc::limit(newspeed, -maxSpeed, maxSpeed);
    
        newturnrate = (r-l);
        newturnrate = PlayerCc::limit(newturnrate, -maxTurnRate, maxTurnRate);
        newturnrate = PlayerCc::dtor(newturnrate);
    
        std::cout << "speed: " << newspeed
                    << "turn: " << newturnrate
                    << std::endl;
    
        if( newspeed < 0.05 && newturnrate < fabs(PlayerCc::dtor(1.0)) )
        {
            // Stalled,  turn 180
            std::cout << "\nStuck! turning 180!!\n";
            
            time_t turnStartTime, turnCurTime;
            time(&turnStartTime);
            time(&turnCurTime);
            
            double diff = 0.0;

	    if( maxTurnRate > 10 )
	    {
            	newturnrate = maxTurnRate;
            	float turntime = 180.0/newturnrate;
            	newturnrate = PlayerCc::dtor(maxTurnRate);
            	while(diff < turntime)
            	{
              		robot->Read();
              		pp->SetSpeed(0, newturnrate);
              		diff = difftime(time(&turnCurTime),turnStartTime);
              		//std::cout << "Time left: " << diff << std::endl;
            	}
	   }
        }
    
        // write commands to robot
        pp->SetSpeed(newspeed, newturnrate);

        // Yield for a bit and let robot do SLAM
        pthread_yield();

    } // end while

    pp->SetSpeed(0, 0);

    free(a);
    return NULL;
}

void *PB_moveTo(void *a)
{
  static const int NUM_ARGS = 3;
  float odom[NUM_ARGS];
  memcpy(odom,a,NUM_ARGS*sizeof(float));

  pp->GoTo(odom[0],odom[1],odom[2]);

  free(a);
  return NULL;
}

int PB_GetOdometry(float *odometry)
{
    // this blocks until new data comes; 10Hz by default
    robot->Read();

    odometry[0] = pp->GetXPos();
    odometry[1] = pp->GetYPos();
    odometry[2] = pp->GetYaw();
    return 1;
}

int PB_GetSpeeds(float *xSpeed, float *ySpeed, float *yawSpeed)
{
    // this blocks until new data comes; 10Hz by default
    robot->Read();

    *xSpeed = pp->GetXSpeed();
    *ySpeed = pp->GetYSpeed();
    *yawSpeed = pp->GetYawSpeed();
    return 1;
}

int PB_GetLaserRanges(float *&ranges)
{
    // this blocks until new data comes; 10Hz by default
    robot->Read();

    uint arrSize = lp->GetCount();
    if(arrSize < 1)
      return arrSize;

    ranges = (float *)calloc(arrSize,sizeof(float));
    for(uint i=0; i < arrSize; ++i)
        ranges[i] = lp->GetRange(i);

    return arrSize;
}

int PB_GetLaserPoints(float *&points)
{
    // this blocks until new data comes; 10Hz by default
    robot->Read();

    uint count = lp->GetCount();
    if(count < 1)
      return count;

    uint arrSize = count*2;
    points = (float *)calloc(arrSize,sizeof(float));
    uint j = 0;
    for(uint i=0; i < arrSize; i+=2)
    {
        player_point_2d_t pt = lp->GetPoint(j);
        points[i] = pt.px;
        points[i+1] = pt.py;
        ++j;
    }
    return arrSize;
}
