/*
**  Matlab client for Powerbot
*/

#include <string.h>
#include "mex.h"
#include "socket_mini.h"
#include "server_mini.h"

static int server_socket=-1;
static int server_connection=-1;
static int server_port=-1;
static char *server_hostname;
/* Sockets */
static struct sockaddr_in server_skname;
static struct sockaddr_in peername;
static int backlog;

/* The SIGTERM handler for shutting down */
void sig_shutdown(int arg)
{
  if (server_connection>0 && server_connection!=server_socket) close(server_connection);
  if (server_socket>0) close(server_socket);
  server_connection=-1;
  server_socket=-1;
}

/*
** Function to check wether an argument is
** a double 2d matrix of given size
** -1 as size means any size
*/
int has_size(const mxArray *arg,int mdims,int *msize)
{
const int *dims;
int k,retval;
  dims = mxGetDimensions(arg);
  retval=1;
  if(mxGetClassID(arg) != mxDOUBLE_CLASS) retval=0;
  if(mxGetNumberOfDimensions(arg) != mdims) retval=0;
  for(k=0;k<mdims;k++) {
    if((msize[k] != -1) && (dims[k] != msize[k])) retval=0;
  }
  return retval;
}

void request_connect(char *hostname,int server_port) {
  if(server_connection != -1) {
    mexErrMsgTxt("Already connected.");
  } else {
    server_socket=establish_remote_socket(server_port, &server_skname,
					  hostname);
    if(server_socket==-1) {
      mexErrMsgTxt("Connect failed.");
    } else {
      mexPrintf("Connected to %s:%d\n",hostname,server_port);
    }
    server_connection=server_socket;
  }
}

void request_disconnect() {
  Message *request;
  if(server_connection==-1) {
    mexErrMsgTxt("Already disconnected.");
  } else {
    request=Message_request("q\n");
    Message_send(request,server_connection);
    Message_free(request);
    sig_shutdown(0);
    mexPrintf("Closed connection to server.\n");
  }
}

void request_setpos(float *pan,float *tilt) {
  Message *request,*response;
  /* Set pos */
  mexPrintf("Setpos %g,%g requested.\n",pan[0],tilt[0]);
  request=Message_posn((double)pan[0],(double)tilt[0]);
  Message_send(request,server_connection);
  Message_free(request);
  if(NULL != (response=read_server_command(server_connection))) {
    sscanf((char *)(response->data),"POS:%g:%g\n",pan,tilt);
    Message_free(response);
  } else {
    sig_shutdown(0);
    mexPrintf("Closed connection to server.\n");
  }
}

void request_getpos(float *pan,float *tilt) {
  Message *request,*response;
  /* Get pos */
  mexPrintf("Getpos requested.\n");
  request=Message_request("r\n");
  Message_send(request,server_connection);
  Message_free(request);
  if(NULL != (response=read_server_command(server_connection))) {
    sscanf((char *)(response->data),"POS:%g:%g\n",pan,tilt);
    Message_free(response);
  } else {
    sig_shutdown(0);
    mexPrintf("Closed connection to server.\n");
  }
}

void request_getodom(float *odom_x,float *odom_y,float *odom_yaw){
  Message *request,*response;
  /* Get odom */
  request=Message_request("o\n");
  Message_send(request,server_connection);
  Message_free(request);
  if(NULL != (response=read_server_command(server_connection))) {
    sscanf((char *)(response->data),"ODOM:%g:%g:%g\n",odom_x,odom_y,odom_yaw);
    Message_free(response);
  } else {
    sig_shutdown(0);
    mexPrintf("Closed connection to server.\n");
  }
}

void request_getspeeds(float *xSpeed,float *ySpeed,float *yawSpeed){
  Message *request,*response;
  /* Get speed */
  request=Message_request("v\n");
  Message_send(request,server_connection);
  Message_free(request);
  if(NULL != (response=read_server_command(server_connection))) {
    sscanf((char *)(response->data),"SPEEDS:%g:%g:%g\n",xSpeed,ySpeed,yawSpeed);
    Message_free(response);
  } else {
    sig_shutdown(0);
    mexPrintf("Closed connection to server.\n");
  }
}

int request_getlaser(float **ranges, float **points){
  Message *request,*response;
  /* Get laser readings */
  request=Message_request("s\n");
  Message_send(request,server_connection);
  Message_free(request);
  /* First see how many readings the server will send us*/
  int rgsSize,ptsSize,numBytes;
  if(NULL != (response=read_fixlength_packet(server_connection))) {
    sscanf((char *)(response->data),"LAS_SIZE:%d %d %d\n",&rgsSize,&ptsSize,&numBytes);
  }
  if(rgsSize < 1) /* no readings */
    return 0;
  
  *ranges = calloc(rgsSize, sizeof(float));
  *points = calloc(ptsSize, sizeof(float));

  /* Now get all the readings */
  if(NULL != (response=read_data_packet(server_connection,numBytes)))
  {
    /* Go through the string of data and get all the readings */
    unsigned char *cur = response->data;
    int i;
    /* Get the ranges */
    for(i=0; i < rgsSize; ++i)
    {
      sscanf((char *)cur,"%g ", (*ranges)+i);
      /* Figure out how many chars to move cur forward */
      cur = (unsigned char *)strchr(cur,' ')+1;
    }
    /* Get the points */
    for(i=0; i < ptsSize; ++i)
    {
      sscanf((char *)cur,"%g ", (*points)+i);
      /* Figure out how many chars to move cur forward */
      cur = (unsigned char *)strchr(cur,' ')+1;
    }

    Message_free(response);
  } else {
    return 0;
  }
  
  return rgsSize;
}

void request_moveRobot(float maxSpeed, float maxTurnRate) {
  Message *request,*response;
  /* Move robot */
  mexPrintf("Robot motion requested.\n");
  char data[MSG_BUFSIZE];
  snprintf((char *)data,MSG_BUFSIZE,"m%g:%g\n",maxSpeed,maxTurnRate);
  request=Message_request(data);
  mexPrintf("Before robot moveto send message.\n");
  Message_send(request,server_connection);
  mexPrintf("After robot moveto send message.\n");
  Message_free(request);
}

void request_moveRobotTo(float *pose) {
  Message *request,*response;
  /* Move robot to */
  mexPrintf("Robot move to requested.\n");
  request=Message_moveTo(pose);
  Message_send(request,server_connection);
  Message_free(request);
}

void request_turnRobot(float yaw) {
  Message *request,*response;
  /* Turn robot */
  mexPrintf("Robot turn requested.\n");
  char data[MSG_BUFSIZE];
  snprintf((char *)data,MSG_BUFSIZE,"t%g\n",yaw);
  request=Message_request(data);
  Message_send(request,server_connection);
  Message_free(request);
}

void request_getres(float *res_pan,float *res_tilt) {
  Message *request,*response;
  /* Get res */
  mexPrintf("Getres requested.\n");
  request=Message_request("d\n");
  Message_send(request,server_connection);
  Message_free(request);
  if(NULL != (response=read_server_command(server_connection))) {
    sscanf((char *)(response->data),"RES:%g:%g\n",res_pan,res_tilt);
    Message_free(response);
  } else {
    sig_shutdown(0);
    mexPrintf("Closed connection to server.\n");
  }
}

void request_getlim(float *min_pan,float *max_pan,float *min_tilt,
		    float *max_tilt) {
  Message *request,*response;
  /* Get limits */
  mexPrintf("Getlim requested.\n");
  request=Message_request("l\n");
  Message_send(request,server_connection);
  Message_free(request);
  if(NULL != (response=read_server_command(server_connection))) {
    sscanf((char *)(response->data),"LIM:%g:%g:%g:%g\n",min_pan,max_pan,
	   min_tilt,max_tilt);
    Message_free(response);
  } else {
    sig_shutdown(0);
    mexPrintf("Closed connection to server.\n");
  }
} 

camera_packet *request_getframe() {
  Message *request;
  camera_packet *cpack;
  /* Get frame */
  mexPrintf("Getframe requested.\n");
  request=Message_request("c\n");
  Message_send(request,server_connection);
/*   Message_printf("sent:",request); */
  Message_free(request);
/*   server_listen(server_connection); */
/*   mexPrintf("After listen\n"); */
  cpack=read_camera_packet(server_connection);
  mexPrintf("After read_camera_packet\n");
  return cpack;
}

void request_setzoom(int zoom) {
  Message *request;
  /* Set zoom */
  mexPrintf("Setzoom %d requested.\n",zoom);
  request=Message_zoom(zoom);
  Message_send(request,server_connection);
  Message_free(request);
}

camera_packet *request_takephoto() {
  Message *request;
  camera_packet *cpack;
  /* Set zoom */
  mexPrintf("take_photo request.\n");
  request=Message_request("x\n");
  Message_send(request,server_connection);
  Message_free(request);
  cpack=read_camera_packet(server_connection);
  mexPrintf("After read_camera_packet\n");
  return cpack;
}

int parsePose(const mxArray *prhs, float *pose)
{
  const int *dims = mxGetDimensions(prhs);
  int ndims = mxGetNumberOfDimensions(prhs);
  if(ndims < 2 || dims[0] != 1 || dims[1] != 3)
  {
    mexPrintf("Error: Expected 3 entries for pose.\n");
    return -1;
  }

  double *matlab_array=mxGetPr(prhs);
  pose[0] = (float)matlab_array[0];
  pose[1] = (float)matlab_array[1];
  pose[2] = (float)matlab_array[2];

  return 0;
}

/*
**  mexFunction is the interface function to MATLAB.
**
** plhs - pointer to left-hand OUTPUT mxArrays
** nlhs - number of left-hand OUTPUT arrays
** prhs - pointer to right-hand INPUT mxArrays
** nrhs - number of right-hand INPUT arrays
**
*/
#define BUF_LEN 50
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[])
{
  const int *arg_dims;
  int arg_ndims;
  int k,l,m;
  int mind,cind,buf_size;
  char cstr[BUF_LEN+1];
  int sport = 7001;  /* Default */
  int zoom = 6;      /* Default */
  float odom_x=0.0,odom_y=0.0,odom_yaw=0.0; /* odometry */
  float xSpeed=0.0,ySpeed=0.0,yawSpeed=0.0; /* speed */
  float *ranges = NULL, *points = NULL; /* laser readings */
  float pose[3]; /* Robot pose */
  float pos_pan=0.0,pos_tilt=0.0;
  float min_pan,max_pan,min_tilt,max_tilt;
  int out_dims[]={0,0,0};
  camera_packet *cpack;
  double *matlab_array;
  unsigned char *matlab_image,*buf;

 /* Argument count check */
  if(nrhs ==0) goto mexErrExit;
    
  /* Parse arg 1 (request string) */
  arg_dims = mxGetDimensions(prhs[0]);
  arg_ndims = mxGetNumberOfDimensions(prhs[0]);
  if(mxGetClassID(prhs[0]) != mxCHAR_CLASS) {
    mexPrintf("Type mismatch: <request> should be of class char.\n");
    goto mexErrExit;
  }
  mxGetString(prhs[0], cstr,BUF_LEN);

  /* CONNECT */
  if(strcmp(cstr,"connect")==0) {
    if((nrhs <2)||(nrhs>3)) {
      mexPrintf("Error: request 'connect' expects 1 or 2 arguments.\n");
      goto mexErrExit;
    }

    if(mxGetClassID(prhs[1]) != mxCHAR_CLASS) {
      mexPrintf("Error: request 'connect' expects a string argument.\n");
      goto mexErrExit;
    }

    mxGetString(prhs[1], cstr,BUF_LEN);
    if(nrhs==3) {
      sport=(int)mxGetScalar(prhs[2]);
    }
    request_connect(cstr,sport);
    /* Remember hostname if successful */
    if(server_socket !=-1) {
      server_hostname=strdup(cstr);
      server_port=sport;
    }
    return;
  }
  
  /* All other requests require an established connection */
  if(server_socket==-1) {
    mexErrMsgTxt("Not connected to server.");
    return;
  }

  /* DISCONNECT */
  if(strcmp(cstr,"disconnect")==0) {
    request_disconnect();
    return;
  }

  /* SETPOS */
  if(strcmp(cstr,"setpos")==0) {
    if((nrhs !=3)) {
      mexPrintf("Error: request 'setpos' expects 2 arguments.\n");
      goto mexErrExit;
    }
    pos_pan=mxGetScalar(prhs[1]);
    pos_tilt=mxGetScalar(prhs[2]);
    request_setpos(&pos_pan,&pos_tilt);
    if(nlhs>0) {
      out_dims[0]=2;out_dims[1]=1;out_dims[2]=0;
      plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
      matlab_array=mxGetPr(plhs[0]);
      matlab_array[0]=(double)pos_pan;
      matlab_array[1]=(double)pos_tilt;
    }
    return;
  }

  /* GETPOS */
  if(strcmp(cstr,"getpos")==0) {
    request_getpos(&pos_pan,&pos_tilt);
/*     mexPrintf("Position %g,%g\n",pos_pan,pos_tilt); */
    out_dims[0]=2;out_dims[1]=1;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
    matlab_array[0]=(double)pos_pan;
    matlab_array[1]=(double)pos_tilt;
    return;
  }

  if(strcmp(cstr,"getodom")==0) {
    request_getodom(&odom_x,&odom_y,&odom_yaw);
    out_dims[0]=1;out_dims[1]=3;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
    matlab_array[0]=(double)odom_x;
    matlab_array[1]=(double)odom_y;
    matlab_array[2]=(double)odom_yaw;
    return;
  }

  if(strcmp(cstr,"getspeeds")==0) {
    request_getspeeds(&xSpeed,&ySpeed,&yawSpeed);
    out_dims[0]=1;out_dims[1]=3;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
    matlab_array[0]=(double)xSpeed;
    matlab_array[1]=(double)ySpeed;
    matlab_array[2]=(double)yawSpeed;
    return;
  }

  if(strcmp(cstr,"getlaser")==0) {
    int numReadings = request_getlaser(&ranges,&points);
    if(numReadings > 0)
    {
      /* return laser ranges */
      out_dims[0]=1;out_dims[1]=numReadings;out_dims[2]=0;
      plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
      matlab_array=mxGetPr(plhs[0]);
      int i;
      for(i=0; i < numReadings; ++i)
        matlab_array[i]=(double)ranges[i];

      /* return laser points */
      out_dims[1]=numReadings*2;
      plhs[1] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
      matlab_array=mxGetPr(plhs[1]);
      for(i=0; i < out_dims[1]; ++i)
        matlab_array[i]=(double)points[i];

      free(ranges);
      free(points);
    } else {
      mexPrintf("Getlaser failed. Disconnecting.\n");
      request_disconnect();
      usleep(100000);
      mexPrintf("Trying to reconnect. %s %d\n",server_hostname,server_port);
      request_connect(server_hostname,server_port);
/*       sig_shutdown(0); */
/*       mexPrintf("Closed connection to server.\n"); */
      out_dims[0]=out_dims[1]=out_dims[2]=0;
      plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
      plhs[1] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    }
    return;
  }

  /* GETFRAME */
  if(strcmp(cstr,"getframe")==0) {
    cpack=request_getframe();
    if(cpack != NULL) {
      mexPrintf("Image size: %d*%d Packet size: %d\n",cpack->width,cpack->height,
		cpack->size);
      if(0) {
	/* Convert RGB image */
	out_dims[0]=cpack->height;
	out_dims[1]=cpack->width;
	out_dims[2]=3;
	plhs[0] = mxCreateNumericArray(3,out_dims,mxUINT8_CLASS,mxREAL);
	matlab_image=(unsigned char *)mxGetPr(plhs[0]);
	for(m=0;m<out_dims[2];m++) {
	  for(l=0;l<out_dims[1];l++) {
	    for(k=0;k<out_dims[0];k++) {
	      mind=k+l*out_dims[0]+m*out_dims[0]*out_dims[1];
	      cind=m+l*out_dims[2]+k*out_dims[2]*out_dims[1];
	      matlab_image[mind]=cpack->data[cind];
	    }
	  }
	}
      } else {
	/* Just output buffer */
	out_dims[0]=cpack->size/cpack->height;/* Extra fields go into width */
	out_dims[1]=cpack->height;
	out_dims[2]=1;
	plhs[0] = mxCreateNumericArray(2,out_dims,mxUINT8_CLASS,mxREAL);
	matlab_image=(unsigned char *)mxGetPr(plhs[0]);
	memcpy(matlab_image,cpack->data,sizeof(char)*cpack->size);     
      }
      free_camera_packet(cpack);
    } else {
      mexPrintf("Getframe failed. Disconnecting.\n");
      request_disconnect();
      usleep(100000);
      mexPrintf("Trying to reconnect. %s %d\n",server_hostname,server_port);
      request_connect(server_hostname,server_port);
/*       sig_shutdown(0); */
/*       mexPrintf("Closed connection to server.\n"); */
      out_dims[0]=out_dims[1]=out_dims[2]=0;
      plhs[0] = mxCreateNumericArray(2,out_dims,mxUINT8_CLASS,mxREAL);      
    }
    return;
  }

  if(strcmp(cstr,"movebot")==0){
    if(nrhs < 3) {
      /* Defaults */
      xSpeed = 0.2;
      yawSpeed = 40.0;
    } else {
      xSpeed = mxGetScalar(prhs[1]);
      yawSpeed = mxGetScalar(prhs[2]);
    }
    request_moveRobot(xSpeed,yawSpeed);
    return;
  }

  if(strcmp(cstr,"movebotTo")==0){
    if(nrhs != 2) {
      mexPrintf("Error: request 'movebotTo' expects 1 argument.\n");
      goto mexErrExit;
    }

    if(parsePose(prhs[1],pose) < 0)
      goto mexErrExit;

    request_moveRobotTo(pose);
    if(nlhs>0) {
      out_dims[0]=1;out_dims[1]=3;out_dims[2]=0;
      plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
      matlab_array=mxGetPr(plhs[0]);
      matlab_array[0]=pose[0];
      matlab_array[1]=pose[1];
      matlab_array[2]=pose[2];
    }
    return;
  }

  if(strcmp(cstr,"turnbot")==0){
    if((nrhs !=2)) {
      mexPrintf("Error: request 'turnbot' expects 1 argument.\n");
      goto mexErrExit;
    }
    odom_yaw = mxGetScalar(prhs[1]);
    request_turnRobot(odom_yaw);
    return;
  }

  /* GETRES */
  if(strcmp(cstr,"getres")==0) {
    request_getres(&pos_pan,&pos_tilt);
/*     mexPrintf("resolution: %g %g\n",pos_pan,pos_tilt); */
    out_dims[0]=2;out_dims[1]=1;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
    matlab_array[0]=(double)pos_pan;
    matlab_array[1]=(double)pos_tilt;
    return;   
  }

  /* GETLIM */
  if(strcmp(cstr,"getlim")==0) {
    request_getlim(&min_pan,&max_pan,&min_tilt,&max_tilt);
/*     mexPrintf("limits: %g %g %g %g\n",min_pan,max_pan,min_tilt,max_tilt); */
    out_dims[0]=4;out_dims[1]=1;out_dims[2]=0;
    plhs[0] = mxCreateNumericArray(2,out_dims,mxDOUBLE_CLASS,mxREAL);
    matlab_array=mxGetPr(plhs[0]);
    matlab_array[0]=(double)min_pan;
    matlab_array[1]=(double)max_pan;
    matlab_array[2]=(double)min_tilt;
    matlab_array[3]=(double)max_tilt;
    return;   
  }

  /* SETZOOM */
  if(strcmp(cstr,"setzoom")==0) {
    if((nrhs >2)||(nrhs <1)) {
      mexPrintf("Error: request 'setzoom' expects one argument.\n");
      goto mexErrExit;
    }
    if(nrhs==2) {
      zoom=mxGetScalar(prhs[1]);
    }
    request_setzoom(zoom);
    return;
  }

  /* TAKE_PHOTO */
  if(strcmp(cstr,"take_photo")==0) {
    cpack = request_takephoto();
    if(cpack != NULL) {
      /* width and height fields cannot be trusted for large images */
      out_dims[0]=cpack->size/64;out_dims[1]=64;
      mexPrintf("dims=[%d,%d] size=%d\n",out_dims[0],out_dims[1],cpack->size);
      plhs[0] = mxCreateNumericArray(2,out_dims,mxUINT8_CLASS,mxREAL);
      memcpy(mxGetPr(plhs[0]),cpack->data,cpack->size);
      free_camera_packet(cpack);
    } else {
      mexPrintf("Getframe failed. Disconnecting.\n");
      request_disconnect();
      usleep(100000);
      mexPrintf("Trying to reconnect. %s %d\n",server_hostname,server_port);
      request_connect(server_hostname,server_port);
      /* Return empty list */
      out_dims[0]=out_dims[1]=out_dims[2]=0;
      plhs[0] = mxCreateNumericArray(2,out_dims,mxUINT8_CLASS,mxREAL);      
    } 
    return;
  }

  mexErrMsgTxt("powerbot: Unknown request");
  goto mexExit;

  /* Exit with error message */
 mexErrExit:
  mexErrMsgTxt("Usage: [<Limg>,<Rimg>]=powerbot(<request>,...)\n");
 mexExit: {}
}
