function [ goal exploreDone ] = globalPlanner(oGrid,objs,planningPars,cameraPars,configPars)
% Choose the next goal
%
% oGrid - The Occupancy grid
% objs - Salient objects
% CAMERAHEIGHT - Height of camera off the ground
% MAX_LOOK_DIST - maximum distance from which to see objects
% probExplore - Probability that we will explore
% explored - binary map of explored frontier points
%

% Unpack parameters
OBSTACLE_THRESHOLD = planningPars.OBSTACLE_THRESHOLD;
explored = planningPars.explored;
probExplore = planningPars.probExplore;
MAX_COST = planningPars.MAX_COST;
if cameraPars.canonfl
  CAMERAHEIGHT = cameraPars.CC_HEIGHT40;
else
    CAMERAHEIGHT = cameraPars.BB_HEIGHT40;
end;
MAX_LOOK_DIST = cameraPars.MAX_LOOK_DIST;

rPose = performSlam('get_curvecpos');
rGridPos = performSlam('get_curgridpos');

exploreDone = false;

% Add visual obstacles
% positionPars.oGrid = oGrid;
% positionPars.rPose = rPose;
% 
% [oGrid nNewObs] = addObstaclesToMap(false,planningPars.OBSTACLE_THRESHOLD,positionPars,cameraPars,configPars);

% Build map of costs to every location
[dists parents nFreeIndices free_indices distMap] = buildPathMap(oGrid,rGridPos,OBSTACLE_THRESHOLD);

exploreFirst = rand(1) < probExplore;
for h=1:2

  if (exploreFirst && h==1 || ~exploreFirst && h==2) && nFreeIndices > 0
 
    fprintf('Attempting to explore...\n');
    % See if we can do exploration
    [frontiers lengths] = getFrontiers(oGrid,floor(255*OBSTACLE_THRESHOLD),20,explored);
    nFrontiers = length(frontiers);
    if nFrontiers > 0
      % Find "best" frontier
      minCost = MAX_COST;
      [height width]=size(oGrid);
      for i=1:length(frontiers)
        frontier = frontiers{i};
        % Find the "best" cell of this frontier
        center = frontier(1,:);

        if norm(center-rGridPos,2) < 20
          continue;
        end;
	    min_dist = inf;
        for j=1:nFreeIndices
          index = free_indices(j);
          thisRow = mod(index-1,height)+1;
          thisCol = floor(index/height)+1;
          thisDist = norm( [ thisRow thisCol ] - center, 2 );

          cost = dists(thisRow,thisCol);
          if( thisDist < min_dist && cost < MAX_COST)
            min_dist = thisDist;
            bestFCell = [ thisRow thisCol ];
            minCost = cost;
            goal.frontier = frontier;
            bestFCenter = center;
          end;
        end;
        
      end; % for i
      if minCost ~= MAX_COST

        goal.path = buildGridPath(parents,rGridPos,bestFCell);
        goal.type = 'e';
        goal.result = bestFCell;
        % Angle to face center of frontier from goal
        y = bestFCell(1)-bestFCenter(1); % high grid row is low y in odometry
        x = bestFCenter(2)-bestFCell(2);
        goal.yaw = mod(atan2(y,x),2*pi);
        return;
      else
        
          goal.type = 'x';
          return
          
        fprintf('No valid path to frontier found, setting best angle.\n');
        goal.type = 'a';
        goal.frontier = frontiers{1};
        % Find the "best" cell of this frontier
        center = goal.frontier(1,:);
        goal.result = center;
        
        goalVec = performSlam('get_vecpos_from_gridpos',center);
        
        x = goalVec(1) - rPose(1);
        y = goalVec(2) - rPose(2);
        
        desiredYaw = atan2(y,x);
        
        % Pass out the relative yaw from the current pose so that this can
        % be applied easily to the odometry
        goal.yaw = desiredYaw - rPose(3);
        return;
      end;
    else
      fprintf('No frontiers to explore\n');
      exploreDone = true;
    end; % if nFrontiers

  elseif exploreFirst && h==2 || ~exploreFirst && h==1

    % Not doing this anymore
    goal.type = 'x';
    return;
      
    fprintf('Attempting to look at previous objects...\n');
    % Find a nearby object with a good viewing angle
    objsSize = size(objs);
    if objsSize(1) > 0

      % 1D Gaussian filter
      sigma = 4;
      gf = exp(-((-sigma*3:sigma*3).^2)/(2*sigma^2));
      gf = gf/sum(gf);

      % Try to find a good viewing angle for each object
      views = [];
      for i=1:objsSize(1)
        object = objs(i,:);
        if object(1) == 0 && object(2) == 0 && object(3) == 0
          % Object with unknown location, skip it
          continue;
        end;

        viewFreqs = object(5:40);

        % Apply Gaussian filter to frequencies
        % Wraparound entries at the edges
        wrappedFreqs = [viewFreqs(37-sigma*3:36) viewFreqs viewFreqs(1:sigma*3)];
        blurredFreqs = conv(gf,wrappedFreqs);
        blurredFreqs = blurredFreqs(sigma*6+1:sigma*6+36);
        if nnz(blurredFreqs < 0.03) == 0
          % Seen enough of this object, skip
          fprintf('Enough views of obj%d,skipping!\n',i);
          continue;
        end;
  
        % Sort angles according to frequencies
        [freqs angles] = sort(blurredFreqs,'ascend');

%         goodAngles = indices(1);
%         denom = sorted(1);
%         if denom == 0
%           denom = 0.001;
%         end;
%         count = 2;
%         for j=2:length(sorted)
%           if sorted(j)/denom < 1.1
%             angles(count) = indices(j);
%             count = count + 1;
%           end;
%         end;
  
        % Matrix to rotate 2d coordinates from robot to global frame
        yaw = rPose(3);
        R_r = [cos(yaw), -sin(yaw); sin(yaw), cos(yaw)];

        % See which view is closest to the robot
        objLoc = [R_r*object(1:2)' + [rPose(1) rPose(2)]';object(3)]';

        % Assume optimal distance is the one that gives 40 degree tilt
        optTilt = pi/4.5;
        optDist = (CAMERAHEIGHT-objLoc(3))*tan(optTilt);
        maxDist = MAX_LOOK_DIST*cos(optTilt);
        if optDist < 0.5
          optDist = 0.5;
        elseif optDist > maxDist
          optDist = maxDist;
        end;

        bestCost = MAX_COST;
        for k=1:length(angles)
          angle = angles(k) *pi/18;

          if freqs(k) > 0.05
            % Done looking at all the good angles for this object
            break;
          end;

          % Find a position from which to see the object at the given angle
          nTries = 1;
          inc = 0;
          reachedMin = 0;
          while nTries < 10
            horzDist = optDist + inc;
            viewLoc = objLoc(1:2) + [horzDist*cos(angle) horzDist*sin(angle)];
            viewGridPos = performSlam('get_gridpos_from_vecpos',viewLoc);
            cost = double(dists(viewGridPos(1), viewGridPos(2)));
            if cost < MAX_COST
              break;
            end;

            % Try +0.2,-0.2,+0.4,-0.4,etc.
            if mod(nTries,2) == 0 && ~reachedMin
              inc = -inc;
            else
              inc = inc + 0.2;
            end;
            if optDist + inc < 0.2
              reachedMin = 1
              inc = -inc + 0.2;
            elseif optDist + inc > maxDist
              break;
            end;

            nTries = nTries + 1;
          end; % while
          dist = norm(viewLoc-rPose(1:2));
          if cost < bestCost && dist > 1.0
            bestCost = cost;
            bestAngle = angle;
            bestView = viewGridPos;
          end;
        end; % for k
  
        if bestCost ~= MAX_COST
          fprintf('Found obj=%d angle=%g cost=%g\n',i,bestAngle,bestCost);
          views = [views; bestCost object(4) i bestView mod(bestAngle+pi,2*pi) objLoc];
        end;
      end; % for i
      sizeViews = size(views);
      if sizeViews(1) > 0
        % Next goal is the position with the lowest cost that gives us a good
        % new view of an object
        [sorted indices] = sortrows(views);
        goal.type = 'o';
        goal.objId = sorted(1,3);
        goal.path = buildGridPath(parents,rGridPos,sorted(1,4:5));
        goal.yaw = mod(sorted(1,6),2*pi);
        goal.result = sorted(1,:);
        return;
      else
        fprintf('No path to good views found!\n');
      end;
    else
      fprintf('Object list is empty!\n');
    end; % if objSize(1) > 0
  end; % Look at objects
end; %for h

% No valid goal found
fprintf('Failed to find a goal\n');
goal.type = 'x';


