function success = movePath(oGrid,planningPars,cameraPars,configPars,goal,logFile)
% Move the robot along a predetermined path. Returns whether
% the robot successfully arrived at the goal
%
% oGrid - The occupancy grid map
% obstacleThreshold - threshold above which grid cells are confidently free
%                     space
% path - a series of positions, expressed in 
%        occupancy grid coordinates
% logFile - file handle of the mapping log file
%

OBSTACLE_THRESHOLD = planningPars.OBSTACLE_THRESHOLD;

success = false;
path = goal.path;
pathSize = size(path);

% tic; commented because there is now a tick in SRVC_Phase2

for i=1:pathSize(1)
    
  % Need current position... do Slam to make sure we have a fresh estimate
  doSlam(1,true,logFile);
  odom = powerbot('getodom');
  rPose = performSlam('get_curvecpos');
  oGrid = performSlam('get_map');
  rGridPos = performSlam('get_curgridpos');
  
  if configPars.DISPLAY
    % Display map and robot
    figure(11);
    colormap(gray(256));
    image(oGrid);
    hold on;
    plot(rGridPos(2),rGridPos(1),'g+');
    plot(path(:,2),path(:,1),'g-');
    plot(path(i,2),path(i,1),'bo');
    drawnow;
  end;
  
  [dists parents nFreeIndices free_indices distMap] = buildPathMap(oGrid,rGridPos,OBSTACLE_THRESHOLD);
  
  fprintf(1,'Current waypoint has dist: %f from obstacles.\n',distMap(path(i,1),path(i,2)));
  
  if distMap(path(i,1),path(i,2)) < 1
      
      fprintf(1,'Failing movepath because goal is too close to an obstacle.\n');
      success = false;
      powerbot('turnbot',0);
      return;
  end;
 
%   if mod(i,5) == 0
%       % Visually detect obstacles in front of us
%       positionPars.oGrid = oGrid;
%       positionPars.rPose = rPose;
%       [oGrid nNewObs] = addObstaclesToMap(false,planningPars.OBSTACLE_THRESHOLD,positionPars,cameraPars,configPars);
%       if nNewObs > 0
%           fprintf('New Obstacles detected!\n');
%           % Plan around visual obstacles if detected
%           [path pathSize] = checkForObstacles(i,path,oGrid,planningPars);
%           if isempty(path)
%               % Ran into a visual obstacle we can't avoid, abort
%               fprintf('Unable to plan around visual obstacle, aborting!\n');
%               powerbot('turnbot',0);
%               return;
%           end;
%       end;
%   end;

  fprintf('moving to [%g %g]\n',path(i,1),path(i,2));

  % Next destination
  nxtMapPos = performSlam('get_vecpos_from_gridpos',path(i,:));

  % Calculate next odometry
  dirVec = nxtMapPos-rPose(1:2);

  if i == pathSize(1)
    yaw = goal.yaw;
  else
    yaw = atan2(dirVec(2),dirVec(1));
    yaw = mod(yaw,2*pi);
  end;

  nxtOdom = double([odom(1:2)+dirVec yaw]);
  
  if i==1 % First waypoint
      % Skip if we're too close to it
      if norm(dirVec) < 0.5
          i=2;
          fprintf('moving to [%g %g]\n',path(i,1),path(i,2));
          nxtMapPos = performSlam('get_vecpos_from_gridpos',path(i,:));
          dirVec = nxtMapPos-rPose(1:2);
          if i == pathSize(1)
              yaw = goal.yaw;
          else
              yaw = atan2(dirVec(2),dirVec(1));
              yaw = mod(yaw,2*pi);
          end;
          nxtOdom = double([odom(1:2)+dirVec yaw]);
      end;
          
      if abs(mod(odom(3)-yaw+pi,2*pi)-pi) > 10*pi/180
          % First turn the robot in place to the correct direction
          angleDiff = (mod(yaw-odom(3)+pi,2*pi)-pi)*180/pi;
          turnbot(angleDiff,logFile);
      end;
  end;

  % Move to the new location
  powerbot('movebotTo',nxtOdom);

  speeds = [0 0 0];
  notYetArrived = 1;
  wpTic = toc;
  iter = 0;
  bestDistRemain = norm(nxtOdom(1:2)-odom(1:2));
  origDistRemain = bestDistRemain;
  noImproveTime = 0;
  while notYetArrived
    iter = iter+1;
    % Dave changed. No more timing out. Let's be careful about our
    % planning, and make sure the robot always gets there!
%     if toc > 30
%       % Taking too long, stuck? Abort
%       fprintf('Taking too long, aborting!\n');
%       powerbot('turnbot',0); % Stop the robot
%       return;
%    end;

    % Do SLAM once in a while if we're not turning too fast
    if abs(speeds(3)) < 0.5 && mod(iter,10)
      doSlam(1,0,logFile);
    end;

    % Clip yaw between 0 and 2pi
    odom = powerbot('getodom');
    speeds = powerbot('getspeeds');

    distRemain = norm(nxtOdom(1:2)-odom(1:2));
    
    if distRemain >= bestDistRemain
        
        noImproveTime = toc - wpTic;
    elseif distRemain < bestDistRemain
        wpTic = toc;
        bestDistRemain = distRemain;
    end;
    
    if noImproveTime > 5 || distRemain > bestDistRemain * 2
        % Have not made progress to the goal for too long, or have moved 
        % too far, this is an
        % unacceptable behaviour for this function, so we'll stop the robot 
        % and abort
        
        powerbot('turnbot',0);
        success = false;
        fprintf(1,'Failed movepath.\n');
        return;
    end;       
    
    angleRemain = abs(thetapp(nxtOdom(3)-odom(3)));
    %fprintf('dist left:%g yaw left:%g\n',distRemain,norm(nxtOdom(3)-odom(3)));
    if i == pathSize(1)
      % Final waypoint, make sure we're there, stopped, and the pose is right
      notYetArrived = distRemain > 0.8 | ...
                      angleRemain > 20*pi/180 | ...
                     abs(speeds(1)) > 0.01 | abs(speeds(3)) > 0.01;
    % Dave again. No timing out!
%    elseif toc-wpTic > 6 % Intermediate waypoint, skip if taking too long
%      notYetArrived = 0;
%     fprintf('Skipping waypoint.\n');
    else
      % Intermediate waypoint, just get close to it
      
      notYetArrived = distRemain > 0.8;
      
    end;
    
    if ~notYetArrived
        fprintf(1,'Arrived at waypoint.\n');
    else
        fprintf(1,'Moving with %f distance and %f angle error.\n',distRemain, angleRemain );
    end;
    
  end;
end;

% Successfully arrived at the goal
fprintf('Successfully arrived at goal\n');
powerbot('turnbot',0); % Stop the robot
success = true;

