%function [sml,tml]=spectral_saliency(imp,plist,resl,favg,savg)
%
% Multi channel spectral saliency based on the CVPR07 paper:
% "Saliency Detection: A Spectral Residual Approach",
% By Xiaodi Hou and Liqing Zhang
%
% IMP    RGB image pyramid as generated by IMAGE_PYRAMID
% PLIST  Projection list Nx3
% RESL   List of resolution levels to use (default 4 5 6)
% FAVG   Fouirer domain averaging filter (default average 3)
% SAVG   Spatial domain averaging filter (default disk 2)
%
% IML    Input image cell array
% SML    Saliency map cell array
% TMP    Computation times array
%
%Per-Erik Forssen, July 2007
function [Iml,Sml,Tml]=spectral_saliency(imp,plist,resl,favg,savg)

if nargin<3,resl=[4 5 6];end
if nargin<4,favg=fspecial('average', 3);end % Fourier domain smoothing
if nargin<5,savg=fspecial('disk', 3);end    % Spatial smoothing

Nr=length(resl);   % Number of resolutions
Np=size(plist,1);  % Number of projections

Iml=cell(Nr,Np);   % Input images
Sml=cell(Nr,Np);   % Saliency maps
Tml=zeros(Nr,Np);  % Timing

for k=1:Nr,
  for l=1:Np,
    
    %tic;% Commented by Dave because there is now a tic in SRVC_Phase2.m
    % Build projection
    im=imp{resl(k)}(:,:,1)*plist(l,1)+...
       imp{resl(k)}(:,:,2)*plist(l,2)+...
       imp{resl(k)}(:,:,3)*plist(l,3);
    Iml{k,l}=im; % Store input image
    
    %% Spectral Residual
    myFFT = fft2(im);
    myLogAmplitude = log(abs(myFFT));
    myPhase = angle(myFFT);
    mySpectralResidual = myLogAmplitude - imfilter(myLogAmplitude, favg, 'replicate');
    saliencyMap = abs(ifft2(exp(mySpectralResidual + i*myPhase))).^2;

    %% After Effect (smoothing of saliency)
    saliencyMap = imfilter(saliencyMap, savg);
    saliencyMap = mat2gray(saliencyMap);
    Sml{k,l}=saliencyMap;   
    t1=toc;
    Tml(k,l)=t1;
  end
end
