/*
 * Accept image information from Matlab and produce the rectified colour images
 *
 *
 * David Meger - March 2007
 */
/*===========================================================================
** System Includes
**=========================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/timeb.h>
#include <time.h>

/*===========================================================================
** Point Grey Research Includes
**=========================================================================*/
#include "triclops.h"
#include "pnmutils.h"

/*===========================================================================
** Matlab Includes
**=========================================================================*/
#include "mex.h"

extern "C" {
#include "matrix.h"
#include "mexutil.h"
}

void write_time_diff(char *strg,clock_t t0,clock_t t1) {
  mexPrintf("%s %g\n",strg,(double)(t1-t0)/CLOCKS_PER_SEC);
}

void report_error(char *msg,TriclopsError te) {
  mexPrintf( "%s : %s\n", triclopsErrorToString(te));
}

/*
**  mexFunction is the interface function to MATLAB.
**
** plhs - pointer to left-hand OUTPUT mxArrays
** nlhs - number of left-hand OUTPUT arrays
** prhs - pointer to right-hand INPUT mxArrays
** nrhs - number of right-hand INPUT arrays
**
*/
#define BUF_LEN 50
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[]) {

  const int *arg_dims;
  int arg_ndims;
  TriclopsContext   triclops;
  TriclopsError     te;
  char config_filename[BUF_LEN+1];
  char default_filename[]="bb_id2210020.config";
  int msize[3],x,y,rows,cols;
  unsigned char *matlab_array,*triclops_array,*red,*green,*blue;
  TriclopsCamera field = TriCam_REFERENCE; /* Default */

  if(nrhs ==0) {
    mexPrintf("Usage: <image_rect>=colour_rectify(<image>[,<config_file>,<field>]) field 1=left,0=right\n");
    return;
  }

  if((nrhs<1)||(nrhs>3)) {
    mexPrintf("Error: 1-3 input arguments expected.\n");
    return;
  }

  /* Parse arg 1 which is the colour image */
  msize[0]=-1;msize[1]=-1;msize[2]=3;
  if(!type_check("image",prhs[0],mxUINT8_CLASS,3,msize))
  {
    mexPrintf("Checking of input image type failed.\n");
    return;
  }
  arg_dims = mxGetDimensions(prhs[0]);
  rows=arg_dims[0];
  cols=arg_dims[1];

  TriclopsInput  triclopsInput;
  /* Fill in stuct fields */
  triclopsInput.nrows=rows;
  triclopsInput.ncols=cols;
  if(1) {
    /* The three colour bands (slightly faster) */
    triclopsInput.inputType=TriInp_RGB;
    triclopsInput.rowinc=cols;
    triclopsInput.u.rgb.red   = calloc(rows*cols,sizeof(unsigned char));
    triclopsInput.u.rgb.green = calloc(rows*cols,sizeof(unsigned char));
    triclopsInput.u.rgb.blue  = calloc(rows*cols,sizeof(unsigned char));
    /* Copy image from matlab */
    red=(unsigned char *)triclopsInput.u.rgb.red;
    green=(unsigned char *)triclopsInput.u.rgb.green;
    blue=(unsigned char *)triclopsInput.u.rgb.blue;
    matlab_array=(unsigned char *)mxGetPr(prhs[0]);
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	red[x+y*cols]=matlab_array[y+x*rows+0*rows*cols];
	green[x+y*cols]=matlab_array[y+x*rows+1*rows*cols];
        blue[x+y*cols]=matlab_array[y+x*rows+2*rows*cols];
      }
    }
  } else {
    /* Packed data */
    triclopsInput.inputType=TriInp_RGB_32BIT_PACKED;
    triclopsInput.rowinc=cols*4;
    triclopsInput.u.rgb32BitPacked.data = 
      calloc(rows*cols*4,sizeof(unsigned char));
    /* Copy image from matlab */
    triclops_array=(unsigned char *)triclopsInput.u.rgb32BitPacked.data;
    matlab_array=(unsigned char *)mxGetPr(prhs[0]);
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	triclops_array[4*(x+y*cols)+3]=matlab_array[y+x*rows+0*rows*cols];
	triclops_array[4*(x+y*cols)+2]=matlab_array[y+x*rows+1*rows*cols];
        triclops_array[4*(x+y*cols)+1]=matlab_array[y+x*rows+2*rows*cols];
      }
    }
  }
  /*mexPrintf("Imported image.\n");*/

  if(nrhs>1) {
    /* Parse arg 2 (config_file) */
    arg_dims = mxGetDimensions(prhs[1]);
    arg_ndims = mxGetNumberOfDimensions(prhs[1]);
    if(mxGetClassID(prhs[1]) != mxCHAR_CLASS) {
      mexPrintf("Type mismatch: <config_file> should be of class char.\n");
      return;
    }
    mxGetString(prhs[1], config_filename,BUF_LEN);
  } else {
    /* Use default filename */
    strcpy(config_filename,default_filename);
  }

  if(nrhs>2) {
    msize[0]=1;msize[1]=1;
    if(!type_check("field",prhs[2],mxDOUBLE_CLASS,2,msize)) {
      return;
    }
    switch((int)mxGetScalar(prhs[2])) {
/* 1 should be left and 0 should be right, but for some reason the code below works */
    case 0:
//       mexPrintf("Received a left image.\n");
      field = TriCam_LEFT;
      break;
    case 1:
//       mexPrintf("Received a right image.\n");
      field = TriCam_RIGHT;
      break;
    default:
//       mexPrintf("Received other field image.\n");
      field = TriCam_REFERENCE;
      break;
    }
  }

  /* mexPrintf("Reading config file:%s\n",config_filename); */

  /* extract Triclops context from file */
  te =  triclopsGetDefaultContextFromFile( &triclops, config_filename);
  if ( te != TriclopsErrorOk ) {
    report_error("TriclopsError: GetDefaultContextFromFile",te);
    return;
  }

  /* Now, the significantly new code. Setup the colour structures, and call
     the colour rectify function. */

  /* Over-ride the resolution set in the config file. */  
  triclopsSetResolution(triclops,rows,cols);

  TriclopsColorImage  colorImage;

  /* The actual rectification call */
  triclopsRectifyColorImage( triclops, field, &triclopsInput, &colorImage );

  /* Final step, copy the image data back into Matlab */
  if(nlhs>0) {
    red=(unsigned char *)colorImage.red;
    green=(unsigned char *)colorImage.green;
    blue=(unsigned char *)colorImage.blue;
    int numDims = 3;
    msize[0]=colorImage.nrows;
    msize[1]=colorImage.ncols;
    msize[2]=3;
    rows=msize[0];
    cols=msize[1];
    plhs[0]=mxCreateNumericArray(numDims,msize,mxUINT8_CLASS,mxREAL);
    matlab_array=(unsigned char *)mxGetPr(plhs[0]);
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
        matlab_array[y+x*rows+0*rows*cols]=red[x+y*cols];
        matlab_array[y+x*rows+1*rows*cols]=green[x+y*cols];
        matlab_array[y+x*rows+2*rows*cols]=blue[x+y*cols];
      }
    }
  }

  /* Release image buffers */
  freeInput(&triclopsInput);

  te = triclopsDestroyContext( triclops );
  if ( te != TriclopsErrorOk ) {
    report_error("TriclopsError: triclopsDestroyContext",te);
    return;
  }
  return;
}
