//=============================================================================
// Copyright  2002 Point Grey Research, Inc. All Rights Reserved.
// 
// This software is the confidential and proprietary information of Point
// Grey Research, Inc. ("Confidential Information").  You shall not
// disclose such Confidential Information and shall use it only in
// accordance with the terms of the license agreement you entered into
// with Point Grey Research Inc.
// 
// PGR MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
// SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE, OR NON-INFRINGEMENT. PGR SHALL NOT BE LIABLE FOR ANY DAMAGES
// SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
// THIS SOFTWARE OR ITS DERIVATIVES.
//
// Digiclops is a registered trademark of Point Grey Research Inc.
//=============================================================================

//=============================================================================
// simplestereo.cpp
//
// This program gives a simple example for opening the Digiclops, grabbing
// stereo images, processing them with the Triclops library, and closing 
// the libraries correctly
//=============================================================================

//=============================================================================
// System Includes
//=============================================================================
#include <stdio.h>
#include <stdlib.h>
#include <sys/timeb.h>
#include <time.h>

//=============================================================================
// Point Grey Research Includes
//=============================================================================
#include "digiclops.h"
#include "triclops.h"


#define _HANDLE_DIGICLOPS_ERROR( functionName ) \
{ \
   if( de != DIGICLOPS_ok ) \
   { \
      printf( \
         "Digiclops Error in %s: %s\n", \
	 functionName, \
	 ::digiclopsErrorToString( de ) ); \
      return -1; \
   }  \
} \
\


int
main( int argc, char** argv )
{
   DigiclopsContext  digiclops;
   DigiclopsError    de;
   TriclopsContext   triclops;
   TriclopsError     te;
   double	     percentRate = 0.0;

   switch( argc )
   {
      default :
         fprintf( stderr, 
                  "Usage : %s [%% of bus bandwidth]\n", 
                  argv[0] );
         return( 1 );
      case 2 :
         percentRate = atof( argv[1] );
         if( ( percentRate <= 0.0 ) || ( percentRate > 100.0 ) )
         {
            fprintf( stderr, 
                     "%s : %% of bus bandwidth must be >0.0 and <=100.0 \n", 
                     argv[0] );
            return( 1 );
         }
         fprintf( stderr, 
                  "%s : Trying %lf %% of bus bandwidth\n", 
                  argv[0], 
                  percentRate );
         break;
      case 1 :
         break; 
   }

   de  = digiclopsCreateContext( &digiclops );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsCreateContext()" );
	 
   // initialize Digiclops device 0
   de  = digiclopsInitialize( digiclops, 0 );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsInitialize()" );

   // extract Triclops context from camera
   de = digiclopsGetTriclopsContextFromCamera( digiclops, &triclops );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsGetTriclopsContextFromCamera()" );

   // start the Digiclops device streaming images 
   de  = digiclopsStart( digiclops );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsStart()" );

   // tell the Digiclops to provide stereo image and the reference image
   de  = digiclopsSetImageTypes( digiclops, STEREO_IMAGE | RIGHT_IMAGE );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsSetImageTypes()" );

   de  = digiclopsSetImageResolution( digiclops, DIGICLOPS_FULL );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsSetImageResolution()" );

   // set the frame rate if asked
   DigiclopsFrameRate frameRate;
   if( percentRate > 0.0 )
   {
      if( percentRate <= 12.0 )
      {
         frameRate = DIGICLOPS_FRAMERATE_012;
      }
      else if( percentRate <= 25.0 )
      {
         frameRate = DIGICLOPS_FRAMERATE_025;
      }
      else if( percentRate <= 50.0 )
      {
         frameRate = DIGICLOPS_FRAMERATE_050;
      }
      else
      {
         frameRate = DIGICLOPS_FRAMERATE_100;
      }
      
      de = digiclopsSetFrameRate( digiclops, frameRate );
      _HANDLE_DIGICLOPS_ERROR( "digiclopsSetFrameRate()" );   
   }

   // retrieve the current frame rate
   de = digiclopsGetFrameRate( digiclops, &frameRate );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsGetFrameRate()" );
   
   switch( frameRate )
   {
      case DIGICLOPS_FRAMERATE_100 :
         percentRate = 100;
         break;
      case DIGICLOPS_FRAMERATE_050 :
         percentRate = 50;
         break;
      case DIGICLOPS_FRAMERATE_025 :
         percentRate = 25;
         break;
      default :  // case DIGICLOPS_FRAMERATE_012 :
         percentRate = 12;
         break;
   }
   printf( "Current bus rate : %lf %%\n",
           percentRate );

   // set the stereo resolution and perform required setup
   printf( "Setting up Triclops library...\n" );

   // set disparity range
   //triclopsSetDisparity( triclops, 2, 40 );
   // set the display mapping
   //triclopsSetTextureValidationMapping( triclops, 0 );
   //triclopsSetUniquenessValidationMapping( triclops, 0 );

   // should use the following line if the Digiclops image resolution is 320x240
   //te  = triclopsSetResolutionAndPrepare( triclops, 240, 320, 240, 320 );
   // should use the following line if the Digiclops image resolution is 640x480
   //te  = triclopsSetResolutionAndPrepare( triclops, 240, 320, 480, 640 );
   te  = triclopsSetResolutionAndPrepare( triclops, 240, 320, 768, 1024 );
   //te  = triclopsSetResolutionAndPrepare( triclops, 768, 1024, 768, 1024 );
   //te  = triclopsSetResolutionAndPrepare( triclops, 480, 640, 768, 1024 );
   //te  = triclopsSetResolutionAndPrepare( triclops, 480, 640, 480, 640 );
   if ( te )
   {
      printf( "TriclopsError: triclopsSetResolutionAndPrepare : %s\n", 
	      triclopsErrorToString( te ) );
      return 1;
   }
   printf( "\tdone\n" );

   // perform a simple profiling
   // grab images for approximately 10 seconds and determine the frame rate

   int  elapsedMillis;
#ifdef WIN32
   struct _timeb start, end;
   _ftime( &start );
#else
   struct timeb start, end;
   ftime( &start );
#endif

   int	 nGrabs = 0;
   printf( "Beginning profiling...\n" );
   for( ;; )
   {
      de = digiclopsGrabImage( digiclops );   
      _HANDLE_DIGICLOPS_ERROR( "digiclopsGrabImage()" );

      TriclopsInput  triclopsInput;
      de = digiclopsExtractTriclopsInput( digiclops, STEREO_IMAGE, &triclopsInput );
      _HANDLE_DIGICLOPS_ERROR( "digiclopsExtractTriclopsInput" );
      
      te = triclopsPreprocess( triclops, &triclopsInput );
      if ( te != TriclopsErrorOk )
      {
	 printf( "TriclopsError: triclopsPreprocess : %s\n", 
	         triclopsErrorToString( te ) );
	 return 1;
      }

      te = triclopsStereo( triclops );
      if ( te != TriclopsErrorOk )
      {
	 printf( "TriclopsError: triclopsStereo : %s\n", 
	         triclopsErrorToString( te ) );
	 return 1;
      }

      nGrabs++;

#ifdef WIN32
      _ftime( &end );
#else
      ftime( &end );
#endif

      elapsedMillis  = end.time * 1000   + end.millitm - 
		       start.time * 1000 - start.millitm;
      if ( elapsedMillis > 10 * 1000 )
	 break;
   }

   printf( "%d frames grabbed in %d milliseconds\n", nGrabs, elapsedMillis );
   printf( " => %f frames per second\n", (float) nGrabs * 1000.0 /(float) elapsedMillis );

   // okay, lets save some of the images to disk
   digiclopsWritePPM( digiclops, RIGHT_IMAGE,  "right.ppm" );
   digiclopsWritePPM( digiclops, STEREO_IMAGE, "stereo.ppm" );
   
   TriclopsImage  triclopsImage;
   triclopsGetImage( triclops, TriImg_DISPARITY, TriCam_REFERENCE, &triclopsImage );
   triclopsSaveImage( &triclopsImage, "disp.pgm" );

   // stop the Digiclops device streaming images 
   de  = digiclopsStop( digiclops );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsStop()" );

   de  = digiclopsDestroyContext( digiclops );
   _HANDLE_DIGICLOPS_ERROR( "digiclopsDestroyContext()" );

   return 0;
}
