/*===========================================================================
** System Includes
**=========================================================================*/
#include <stdio.h>
#include <stdlib.h>

/*===========================================================================
** Point Grey Research Includes
**=========================================================================*/
#include "triclops.h"
#include "pnmutils.h"

/*===========================================================================
** Matlab Includes
**=========================================================================*/
#include "mex.h"

void report_error(char *msg,TriclopsError te) {
  mexPrintf( "%s : %s\n", triclopsErrorToString(te));
}

/*
**  mexFunction is the interface function to MATLAB.
**
** plhs - pointer to left-hand OUTPUT mxArrays
** nlhs - number of left-hand OUTPUT arrays
** prhs - pointer to right-hand INPUT mxArrays
** nrhs - number of right-hand INPUT arrays
**
*/

#define BUF_LEN 50
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[]) {
  const int *arg_dims;
  int arg_ndims;
  TriclopsContext triclops;
  TriclopsError te;

  static const int NUM_COLS = 2;
  double *imgPos_array;
  int nImgPos;

  char config_filename[BUF_LEN+1];
  char default_filename[]="bb_id5020143.config";
  int msize[3],x,y,rows,cols;
  unsigned char *matlab_array,*triclops_array;

  /* Argument count check */
  if(nrhs ==0 || nlhs < 1) {
    mexPrintf("Usage: <pc>=triclops_getXYZ(<img pos>,<8bit disparity map>[,<config_file>])\n");
    return;
  }

  if((nrhs<2)||(nrhs>3)) {
    mexPrintf("Error: Two or three input arguments expected.\n");
    return;
  }

  /* Parse arg 1 (image positions) */
  imgPos_array=mxGetPr(prhs[0]);
  arg_dims = mxGetDimensions(prhs[0]);
  nImgPos=arg_dims[0];
  if(arg_dims[1]!= NUM_COLS)
  {
    mexPrintf("Error: <imgPos> should have %d columns.\n",NUM_COLS);
    return;
  }

  /* Parse arg 2 (disparity map) */
  matlab_array=(unsigned char *)mxGetPr(prhs[1]);
  arg_dims = mxGetDimensions(prhs[1]);
  rows=arg_dims[0];
  cols=arg_dims[1];

  TriclopsImage triclopsImage;
  /* Fill in stuct fields */
  triclopsImage.nrows=rows;
  triclopsImage.ncols=cols;
  triclopsImage.rowinc = cols;
  triclopsImage.data = (unsigned char *)mxCalloc(rows*cols,sizeof(unsigned char));

  triclops_array=triclopsImage.data;
  for(x=0;x<cols;++x)
    for(y=0;y<rows;++y)
      triclops_array[x+y*cols] = matlab_array[y+x*rows];

  if(nrhs>1) {
    /* Parse arg 3 (config_file) */
    arg_dims = mxGetDimensions(prhs[2]);
    arg_ndims = mxGetNumberOfDimensions(prhs[2]);
    if(mxGetClassID(prhs[2]) != mxCHAR_CLASS) {
      mexPrintf("Type mismatch: <config_file> should be of class char.\n");
      return;
    }
    mxGetString(prhs[2], config_filename,BUF_LEN);
  } else {
    /* Use default filename */
    strcpy(config_filename,default_filename);
  }
  //  mexPrintf("Reading config file:%s\n",config_filename);

  /* extract Triclops context from file */
  te =  triclopsGetDefaultContextFromFile( &triclops, config_filename);
  if ( te != TriclopsErrorOk ) {
    report_error("TriclopsError: GetDefaultContextFromFile",te);
    return;
  }  

  unsigned char disparity;
  static const int NUM_ENTRIES = 3;

  // Create the returned matlab array
  plhs[0]=mxCreateNumericMatrix(nImgPos,NUM_ENTRIES,mxDOUBLE_CLASS,mxREAL);
  if(plhs[0] == NULL)
  {
    printf("plhs is NULL!\n");
    return;
  }
  double *retArray=mxGetPr(plhs[0]);

  for(int i=0; i<nImgPos; i++)
  {
    float x_w=-1, y_w=-1, z_w=-1; // tombstone values to indicate we failed to get XYZ
    int imgPosRow = (int)imgPos_array[i];
    int imgPosCol = (int)imgPos_array[nImgPos+i];
    disparity = triclops_array[imgPosCol+imgPosRow*triclopsImage.ncols];

    // filter invalid points
    if ( disparity < 252 && disparity > 6 )
    {
      // convert the disparity value to floating point x_w,y_w,z_w
      te = triclopsRCD8ToXYZ( triclops, imgPosRow, imgPosCol, disparity, &x_w, &y_w, &z_w );
      if( te != TriclopsErrorOk )
      {
        printf("TriclopsError: triclopsRCD8ToXYZ\n");
      }
    }

    //printf("disparity=%d %d %d %g %g %g\n",disparity,imgPosRow,imgPosCol,x_w,y_w,z_w);
    // Add [x y z] to returned matlab array
    retArray[i] = x_w;
    retArray[i+nImgPos] = y_w;
    retArray[i+nImgPos*2] = z_w;
  }

  te = triclopsDestroyContext( triclops );
  if ( te != TriclopsErrorOk )
  {
    report_error("TriclopsError: triclopsDestroyContext",te);
  }
}
