/*
 * Read triclops parameters and a stereo pair from file,
 * compute disparity and write to pgm.
 *
 * Per-Erik Forssen, Feb 2007
 */
/*===========================================================================
** System Includes
**=========================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/timeb.h>
#include <time.h>

/*===========================================================================
** Point Grey Research Includes
**=========================================================================*/
#include "triclops.h"
#include "pnmutils.h"

/*===========================================================================
** Matlab Includes
**=========================================================================*/
#include "mex.h"

extern "C" {
#include "mexutil.h"
}

void write_time_diff(char *strg,clock_t t0,clock_t t1) {
  mexPrintf("%s %g\n",strg,(double)(t1-t0)/CLOCKS_PER_SEC);
}

void report_error(char *msg,TriclopsError te) {
  mexPrintf( "%s : %s\n", triclopsErrorToString(te));
}

/*
**  mexFunction is the interface function to MATLAB.
**
** plhs - pointer to left-hand OUTPUT mxArrays
** nlhs - number of left-hand OUTPUT arrays
** prhs - pointer to right-hand INPUT mxArrays
** nrhs - number of right-hand INPUT arrays
**
*/

#define BUF_LEN 50
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[]) {

  const int *arg_dims;
  int arg_ndims;
  TriclopsContext   triclops;
  TriclopsError     te;
  char config_filename[BUF_LEN+1];
  char default_filename[]="bb_id2210020.config";
  int msize[3],x,y,rows,cols;
  unsigned char *matlab_array,*triclops_array,*red,*green,*blue;
  clock_t t0,t1,t2,t3,t4;
  
  t0=clock();
   /* Argument count check */
  if(nrhs ==0) {
    mexPrintf("Usage: <disp>=triclops_stereo(<stereo>[,<config_file>])\n");
    return;
  }

  if((nrhs<1)||(nrhs>2)) {
    mexPrintf("Error: One or two input arguments expected.\n");
    return;
  }

  /* Parse arg 1 (stereo) */
  msize[0]=-1;msize[1]=-1;msize[2]=3;
  if(!type_check("stereo",prhs[0],mxUINT8_CLASS,3,msize))
    return;
  arg_dims = mxGetDimensions(prhs[0]);
  rows=arg_dims[0];
  cols=arg_dims[1];

  TriclopsInput  triclopsInput;
  /* Fill in stuct fields */
  triclopsInput.nrows=rows;
  triclopsInput.ncols=cols;
  if(1) {
    /* The three colour bands (slightly faster) */
    triclopsInput.inputType=TriInp_RGB;
    triclopsInput.rowinc=cols;
    triclopsInput.u.rgb.red   = calloc(rows*cols,sizeof(unsigned char));
    triclopsInput.u.rgb.green = calloc(rows*cols,sizeof(unsigned char));
    triclopsInput.u.rgb.blue  = calloc(rows*cols,sizeof(unsigned char));
    /* Copy image from matlab */
    red=(unsigned char *)triclopsInput.u.rgb.red;
    green=(unsigned char *)triclopsInput.u.rgb.green;
    matlab_array=(unsigned char *)mxGetPr(prhs[0]);
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	red[x+y*cols]=matlab_array[y+x*rows];
	green[x+y*cols]=matlab_array[y+x*rows+rows*cols];
      }
    }
  } else {
    /* Packed data */
    triclopsInput.inputType=TriInp_RGB_32BIT_PACKED;
    triclopsInput.rowinc=cols*4;
    triclopsInput.u.rgb32BitPacked.data = 
      calloc(rows*cols*4,sizeof(unsigned char));
    /* Copy image from matlab */
    triclops_array=(unsigned char *)triclopsInput.u.rgb32BitPacked.data;
    matlab_array=(unsigned char *)mxGetPr(prhs[0]);
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	triclops_array[x*4+y*cols*4+2]=matlab_array[y+x*rows];
	triclops_array[x*4+y*cols*4+1]=matlab_array[y+x*rows+rows*cols];
      }
    }
  }
  //  mexPrintf("Imported image.\n");

  if(nrhs>1) {
    /* Parse arg 2 (config_file) */
    arg_dims = mxGetDimensions(prhs[1]);
    arg_ndims = mxGetNumberOfDimensions(prhs[1]);
    if(mxGetClassID(prhs[1]) != mxCHAR_CLASS) {
      mexPrintf("Type mismatch: <config_file> should be of class char.\n");
      return;
    }
    mxGetString(prhs[1], config_filename,BUF_LEN);
  } else {
    /* Use default filename */
    strcpy(config_filename,default_filename);
  }
  //  mexPrintf("Reading config file:%s\n",config_filename);

  /* extract Triclops context from file */
  te =  triclopsGetDefaultContextFromFile( &triclops, config_filename);
  if ( te != TriclopsErrorOk ) {
    report_error("TriclopsError: GetDefaultContextFromFile",te);
    return;
  }
  t1=clock();

  //  mexPrintf("Doing stereo 1\n");
  te = triclopsPreprocess( triclops, &triclopsInput );
  if ( te != TriclopsErrorOk ) {
    report_error("TriclopsError: triclopsPreprocess",te);
    return;
  }
  t2=clock();
  //  mexPrintf("Doing stereo 2\n");
  te = triclopsStereo( triclops );
  if ( te != TriclopsErrorOk ) {
    report_error("TriclopsError: triclopsStereo",te);
    return;
  }
  t3=clock();

  TriclopsImage  triclopsImage;
  TriclopsImage16 depthImage16;
  triclopsGetImage( triclops, TriImg_DISPARITY, TriCam_REFERENCE,
		    &triclopsImage );
  triclopsGetImage16( triclops, TriImg16_DISPARITY, TriCam_REFERENCE,
            &depthImage16 );
  //   triclopsSaveImage( &triclopsImage, "disp.pgm" ); 

  //  mexPrintf("Doing stereo 3\n");

  if(nlhs>0) {
    msize[0]=triclopsImage.nrows;
    msize[1]=triclopsImage.ncols;
    plhs[0]=mxCreateNumericMatrix(msize[0],msize[1],mxUINT8_CLASS,mxREAL);
    matlab_array=(unsigned char *)mxGetPr(plhs[0]);
    triclops_array=triclopsImage.data;
    for(x=0;x<msize[1];x++) {
      for(y=0;y<msize[0];y++) {
	matlab_array[y+x*msize[0]]=triclops_array[x+y*msize[1]];
      }
    }
  }
  //  mexPrintf("Copied image to matlab\n");

  /* Release image buffers */
  freeInput(&triclopsInput);
  t4=clock();
  //  mexPrintf("Released buffers\n");

//   write_time_diff("setup",t0,t1);
//   write_time_diff("rectification",t1,t2);
//   write_time_diff("stereo",t2,t3);
//   write_time_diff("data copy",t3,t4);
  te = triclopsDestroyContext( triclops );
  if ( te != TriclopsErrorOk ) {
    report_error("TriclopsError: triclopsDestroyContext",te);
    return;
  }
  return;
}
