/*
**  bayer.c
**
**  Routines for interpolating a Bayer-sensor pattern to
**  a Matlab RGB image.
*/
#include <stdlib.h>
#include <math.h>
#include "bayer.h"

/*
 * Write difference between two clock_t structs
 * to standard output.
 *
 *  strg   Message preceeding time text
 *  t0     start time
 *  t1     end time
 *
 */
void write_time_diff(char *strg,clock_t t0,clock_t t1) {
 printf("%s %g\n",strg,(double)(t1-t0)/CLOCKS_PER_SEC);
}
/*
 * Convert a raw Bayer buffer to an RGB buffer
 * by interpolation with an [1 2 1]/2 kernel along
 * each dimension. Data is also scaled by 1.0/255.0
 *
 *  field_ptr pointer to image buffer (assumed to be
 *            the same size as bf_rgb)
 *  bf_rgb    pointer to new rgb buffer
 *  bpattern  Bayer pattern type
 *
 */
void bayer_to_rgb(unsigned char *field_ptr,buffer *bf_rgb,int bpattern) {
  int rows,cols,x,y,xoff=0,yoff=0;
  int cind,red,green,blue,tcomp=BAYER_RGGB;
  int rs,gs,bs,nr,ng,nb;
  int emark;
  fpnum *rgb;

  rows=bf_rgb->rows;
  cols=bf_rgb->cols;
  /* Get pointer to RGB buffer  */
  rgb=bf_rgb->data;

  red=0;
  green=rows*cols;
  blue=2*rows*cols;

  switch(bpattern) {
  case BAYER_BGGR:
    xoff=1;yoff=1;
    break;
  case BAYER_GRBG:
    xoff=1;yoff=0;
    break;
  case BAYER_RGGB:
    xoff=0;yoff=0;
    break;
  case BAYER_GBRG:
    xoff=0;yoff=1;
    break;
  }
  for(x=0;x<cols;x++) {
    for(y=0;y<rows;y++) {      
      cind=y+x*rows;
      
      /* emark flags for edges */
      /*
       *       5 4 6
       *       1 0 2
       *       9 8 A
       */
      emark=0;
      if(x==0)      emark+=1;
      if(x==cols-1) emark+=2;
      if(y==0)      emark+=4;
      if(y==rows-1) emark+=8;
      
      /* Determine which pixel we're at right now */
      tcomp=((x+xoff)%2)+2*((y+yoff)%2);
      switch(tcomp) {
      case 0:
	/* At the RED pixel */
	gs=bs=0;
	rgb[cind+red]=field_ptr[cind]/255.0;      /* Red */
	if(emark==0) {
	  gs+=field_ptr[cind-rows];
	  gs+=field_ptr[cind+rows];
	  gs+=field_ptr[cind-1];
	  gs+=field_ptr[cind+1];
	  rgb[cind+green]=gs/255.0/4.0;           /* Green */
	  bs+=field_ptr[cind-rows-1];
	  bs+=field_ptr[cind-rows+1];
	  bs+=field_ptr[cind+rows-1];
	  bs+=field_ptr[cind+rows+1];
	  rgb[cind+blue]=bs/255.0/4.0;            /* Blue */
	} else {
	  ng=nb=0;
	  if(!(emark&1))  {gs+=field_ptr[cind-rows];ng++;}
	  if(!(emark&2))  {gs+=field_ptr[cind+rows];ng++;}
	  if(!(emark&4))  {gs+=field_ptr[cind-1];ng++;}
	  if(!(emark&8))  {gs+=field_ptr[cind+1];ng++;}
	  rgb[cind+green]=gs/255.0/ng;              /* Green */
	  if(!(emark&5))  {bs+=field_ptr[cind-rows-1];nb++;}
	  if(!(emark&9))  {bs+=field_ptr[cind-rows+1];nb++;}
	  if(!(emark&6))  {bs+=field_ptr[cind+rows-1];nb++;}
	  if(!(emark&10)) {bs+=field_ptr[cind+rows+1];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;               /* Blue */
	}
	break;
      case 1:
	/* At the GREEN1 pixel */
	rs=bs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind-rows];
	  rs+=field_ptr[cind+rows];
	  rgb[cind+red]=rs/255.0/2.0;               /* Red */
	  gs+=field_ptr[cind-1-rows];
	  gs+=field_ptr[cind-1+rows];
	  gs+=4.0*field_ptr[cind];
	  gs+=field_ptr[cind+1-rows];
	  gs+=field_ptr[cind+1+rows];
	  rgb[cind+green]=gs/255.0/8.0;             /* Green */
	  bs+=field_ptr[cind-1];
	  bs+=field_ptr[cind+1];
	  rgb[cind+blue]=bs/255.0/2.0;              /* Blue */
	} else {
	  nr=nb=ng=0;
	  if(!(emark&1))  {rs+=field_ptr[cind-rows];nr++;}
	  if(!(emark&2))  {rs+=field_ptr[cind+rows];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&5))  {gs+=field_ptr[cind-1-rows];ng++;}
	  if(!(emark&6))  {gs+=field_ptr[cind-1+rows];ng++;}
	  gs+=4.0*field_ptr[cind];ng+=4;
	  if(!(emark&9))  {gs+=field_ptr[cind+1-rows];ng++;}
	  if(!(emark&10)) {gs+=field_ptr[cind+1+rows];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  if(!(emark&4))  {bs+=field_ptr[cind-1];nb++;}
	  if(!(emark&8))  {bs+=field_ptr[cind+1];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;              /* Blue */
	}
	break;
      case 2:
	/* At the GREEN2 pixel */
	rs=bs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind-1];
	  rs+=field_ptr[cind+1];
	  rgb[cind+red]=rs/255.0/2.0;               /* Red */
	  gs+=field_ptr[cind-1-rows];
	  gs+=field_ptr[cind-1+rows];
	  gs+=4.0*field_ptr[cind];
	  gs+=field_ptr[cind+1-rows];
	  gs+=field_ptr[cind+1+rows];
	  rgb[cind+green]=gs/255.0/8.0;             /* Green */
	  bs+=field_ptr[cind-rows];
	  bs+=field_ptr[cind+rows];
	  rgb[cind+blue]=bs/255.0/2.0;              /* Blue */
	} else {
	  nr=nb=ng=0;
	  if(!(emark&4))  {rs+=field_ptr[cind-1];nr++;}
	  if(!(emark&8))  {rs+=field_ptr[cind+1];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&5))  {gs+=field_ptr[cind-1-rows];ng++;}
	  if(!(emark&6))  {gs+=field_ptr[cind-1+rows];ng++;}
	  gs+=4.0*field_ptr[cind];ng+=4;
	  if(!(emark&9))  {gs+=field_ptr[cind+1-rows];ng++;}
	  if(!(emark&10)) {gs+=field_ptr[cind+1+rows];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  if(!(emark&1))  {bs+=field_ptr[cind-rows];nb++;}
	  if(!(emark&2))  {bs+=field_ptr[cind+rows];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;              /* Blue */
	}
	  break;
      case 3:
	/* At the BLUE pixel */
	rs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind-rows-1];
	  rs+=field_ptr[cind-rows+1];
	  rs+=field_ptr[cind+rows-1];
	  rs+=field_ptr[cind+rows+1];
	  rgb[cind+red]=rs/255.0/4.0;           /* Red */
	  gs+=field_ptr[cind-rows];
	  gs+=field_ptr[cind+rows];
	  gs+=field_ptr[cind-1];
	  gs+=field_ptr[cind+1];
	  rgb[cind+green]=gs/255.0/4.0;         /* Green */
	  rgb[cind+blue]=field_ptr[cind]/255.0; /* Blue */
	} else {
	  nr=ng=0;
	  if(!(emark&5))  {rs+=field_ptr[cind-rows-1];nr++;}
	  if(!(emark&9))  {rs+=field_ptr[cind-rows+1];nr++;}
	  if(!(emark&6))  {rs+=field_ptr[cind+rows-1];nr++;}
	  if(!(emark&10)) {rs+=field_ptr[cind+rows+1];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&1))  {gs+=field_ptr[cind-rows];ng++;}
	  if(!(emark&2))  {gs+=field_ptr[cind+rows];ng++;}
	  if(!(emark&4))  {gs+=field_ptr[cind-1];ng++;}
	  if(!(emark&8))  {gs+=field_ptr[cind+1];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  rgb[cind+blue]=field_ptr[cind]/255.0;    /* Blue */
	}
	break;
      }
    }
  }
}
/*
 * Convert a field of a raw Bayer stereo buffer to an RGB buffer
 * by interpolation with an [1 2 1]/2 kernel along
 * each dimension. Data is also scaled by 1.0/255.0
 * This function also transposes input and output images.
 *
 *  field_ptr pointer to image buffer (assumed to be
 *            the same size as bf_rgb)
 *  bf_rgb    pointer to new rgb buffer
 *  bpattern  Bayer pattern type
 *  field     Field to extract, 0/1 for left/right
 *
 */
void bayer_to_rgb_stereo16(unsigned char *field_ptr,buffer *bf_rgb,int bpattern,int field) {
  int rows,cols,x,y,xoff=0,yoff=0;
  int cind,cind2,red,green,blue,tcomp=BAYER_RGGB;
  int rs,gs,bs,nr,ng,nb;
  int emark;
  fpnum *rgb;

  rows=bf_rgb->rows;
  cols=bf_rgb->cols;
  /* Get pointer to RGB buffer  */
  rgb=bf_rgb->data;

  red=0;
  green=rows*cols;
  blue=2*rows*cols;

  switch(bpattern) {
  case BAYER_BGGR:
    yoff=1;xoff=1;
    break;
  case BAYER_GRBG:
    yoff=1;xoff=0;
    break;
  case BAYER_RGGB:
    yoff=0;xoff=0;
    break;
  case BAYER_GBRG:
    yoff=0;xoff=1;
    break;
  }
  for(x=0;x<cols;x++) {
    for(y=0;y<rows;y++) {      
      cind=y+x*rows;
      cind2=x*2+y*cols*2+field;
      
      /* emark flags for edges */
      /*
       *       5 4 6
       *       1 0 2
       *       9 8 A
       */
      emark=0;
      if(y==0)      emark+=1;
      if(y==rows-1) emark+=2;
      if(x==0)      emark+=4;
      if(x==cols-1) emark+=8;
      
      /* Determine which pixel we're at right now */
      tcomp=((y+yoff)%2)+2*((x+xoff)%2);
      switch(tcomp) {
      case 0:
	/* At the RED pixel */
	gs=bs=0;
	rgb[cind+red]=field_ptr[cind2]/255.0;      /* Red */
	if(emark==0) {
	  gs+=field_ptr[cind2-2*cols];
	  gs+=field_ptr[cind2+2*cols];
	  gs+=field_ptr[cind2-2];
	  gs+=field_ptr[cind2+2];
	  rgb[cind+green]=gs/255.0/4.0;           /* Green */
	  bs+=field_ptr[cind2-2*cols-2];
	  bs+=field_ptr[cind2-2*cols+2];
	  bs+=field_ptr[cind2+2*cols-2];
	  bs+=field_ptr[cind2+2*cols+2];
	  rgb[cind+blue]=bs/255.0/4.0;            /* Blue */
	} else {
	  ng=nb=0;
	  if(!(emark&1))  {gs+=field_ptr[cind2-2*cols];ng++;}
	  if(!(emark&2))  {gs+=field_ptr[cind2+2*cols];ng++;}
	  if(!(emark&4))  {gs+=field_ptr[cind2-2];ng++;}
	  if(!(emark&8))  {gs+=field_ptr[cind2+2];ng++;}
	  rgb[cind+green]=gs/255.0/ng;              /* Green */
	  if(!(emark&5))  {bs+=field_ptr[cind2-2*cols-2];nb++;}
	  if(!(emark&9))  {bs+=field_ptr[cind2-2*cols+2];nb++;}
	  if(!(emark&6))  {bs+=field_ptr[cind+2*cols-2];nb++;}
	  if(!(emark&10)) {bs+=field_ptr[cind+2*cols+2];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;               /* Blue */
	}
	break;
      case 1:
	/* At the GREEN1 pixel */
	rs=bs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind2-2*cols];
	  rs+=field_ptr[cind2+2*cols];
	  rgb[cind+red]=rs/255.0/2.0;               /* Red */
	  gs+=field_ptr[cind2-2-2*cols];
	  gs+=field_ptr[cind2-2+2*cols];
	  gs+=4.0*field_ptr[cind2];
	  gs+=field_ptr[cind2+2-2*cols];
	  gs+=field_ptr[cind2+2+2*cols];
	  rgb[cind+green]=gs/255.0/8.0;             /* Green */
	  bs+=field_ptr[cind2-2];
	  bs+=field_ptr[cind2+2];
	  rgb[cind+blue]=bs/255.0/2.0;              /* Blue */
	} else {
	  nr=nb=ng=0;
	  if(!(emark&1))  {rs+=field_ptr[cind2-2*cols];nr++;}
	  if(!(emark&2))  {rs+=field_ptr[cind2+2*cols];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&5))  {gs+=field_ptr[cind2-2-2*cols];ng++;}
	  if(!(emark&6))  {gs+=field_ptr[cind2-2+2*cols];ng++;}
	  gs+=4.0*field_ptr[cind2];ng+=4;
	  if(!(emark&9))  {gs+=field_ptr[cind2+2-2*cols];ng++;}
	  if(!(emark&10)) {gs+=field_ptr[cind2+2+2*cols];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  if(!(emark&4))  {bs+=field_ptr[cind2-2];nb++;}
	  if(!(emark&8))  {bs+=field_ptr[cind2+2];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;              /* Blue */
	}
	break;
      case 2:
	/* At the GREEN2 pixel */
	rs=bs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind2-2];
	  rs+=field_ptr[cind2+2];
	  rgb[cind+red]=rs/255.0/2.0;               /* Red */
	  gs+=field_ptr[cind2-2-2*cols];
	  gs+=field_ptr[cind2-2+2*cols];
	  gs+=4.0*field_ptr[cind2];
	  gs+=field_ptr[cind2+2-2*cols];
	  gs+=field_ptr[cind2+2+2*cols];
	  rgb[cind+green]=gs/255.0/8.0;             /* Green */
	  bs+=field_ptr[cind2-2*cols];
	  bs+=field_ptr[cind2+2*cols];
	  rgb[cind+blue]=bs/255.0/2.0;              /* Blue */
	} else {
	  nr=nb=ng=0;
	  if(!(emark&4))  {rs+=field_ptr[cind2-2];nr++;}
	  if(!(emark&8))  {rs+=field_ptr[cind2+2];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&5))  {gs+=field_ptr[cind2-2-2*cols];ng++;}
	  if(!(emark&6))  {gs+=field_ptr[cind2-2+2*cols];ng++;}
	  gs+=4.0*field_ptr[cind2];ng+=4;
	  if(!(emark&9))  {gs+=field_ptr[cind2+2-2*cols];ng++;}
	  if(!(emark&10)) {gs+=field_ptr[cind2+2+2*cols];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  if(!(emark&1))  {bs+=field_ptr[cind2-2*cols];nb++;}
	  if(!(emark&2))  {bs+=field_ptr[cind2+2*cols];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;              /* Blue */
	}
	  break;
      case 3:
	/* At the BLUE pixel */
	rs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind2-2*cols-2];
	  rs+=field_ptr[cind2-2*cols+2];
	  rs+=field_ptr[cind2+2*cols-2];
	  rs+=field_ptr[cind2+2*cols+2];
	  rgb[cind+red]=rs/255.0/4.0;           /* Red */
	  gs+=field_ptr[cind2-2*cols];
	  gs+=field_ptr[cind2+2*cols];
	  gs+=field_ptr[cind2-2];
	  gs+=field_ptr[cind2+2];
	  rgb[cind+green]=gs/255.0/4.0;         /* Green */
	  rgb[cind+blue]=field_ptr[cind2]/255.0; /* Blue */
	} else {
	  nr=ng=0;
	  if(!(emark&5))  {rs+=field_ptr[cind2-2*cols-2];nr++;}
	  if(!(emark&9))  {rs+=field_ptr[cind2-2*cols+2];nr++;}
	  if(!(emark&6))  {rs+=field_ptr[cind2+2*cols-2];nr++;}
	  if(!(emark&10)) {rs+=field_ptr[cind2+2*cols+2];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&1))  {gs+=field_ptr[cind2-2*cols];ng++;}
	  if(!(emark&2))  {gs+=field_ptr[cind2+2*cols];ng++;}
	  if(!(emark&4))  {gs+=field_ptr[cind2-2];ng++;}
	  if(!(emark&8))  {gs+=field_ptr[cind2+2];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  rgb[cind+blue]=field_ptr[cind2]/255.0;    /* Blue */
	}
	break;
      }
    }
  }
}
/*
 * Convert a field of a possibly multi-field raw Bayer to
 * an RGB buffer by interpolation with an [1 2 1]/2 kernel along
 * each dimension. Data is also scaled by 1.0/255.0
 * This function also transposes input and output images.
 *
 *  field_ptr pointer to image buffer (assumed to be
 *            the same size as bf_rgb)
 *  bf_rgb    pointer to new rgb buffer
 *  bpattern  Bayer pattern type
 *  fields    Number of fields
 *  field     Field to extract, 0/1 for left/right
 *
 */
void bayer_to_rgb_fields(unsigned char *field_ptr,buffer *bf_rgb,int bpattern,int fields,int field) {
  int rows,cols,x,y,xoff=0,yoff=0;
  int cind,cind2,red,green,blue,tcomp=BAYER_RGGB;
  int bayer_r,bayer_d,bayer_rd,bayer_ld;
  int rs,gs,bs,nr,ng,nb;
  int emark;
  fpnum *rgb;

  rows=bf_rgb->rows;
  cols=bf_rgb->cols;
  /* Get pointer to RGB buffer  */
  rgb=bf_rgb->data;

  /* Offsets */
  red=0;
  green=rows*cols;
  blue=2*rows*cols;
  
  /* Bayer offsets:   */
  bayer_r  = fields;
  bayer_d  = fields*cols;
  bayer_rd = fields*cols+fields;
  bayer_ld = fields*cols-fields;

  switch(bpattern) {
  case BAYER_BGGR:
    yoff=1;xoff=1;
    break;
  case BAYER_GRBG:
    yoff=1;xoff=0;
    break;
  case BAYER_RGGB:
    yoff=0;xoff=0;
    break;
  case BAYER_GBRG:
    yoff=0;xoff=1;
    break;
  }
  for(x=0;x<cols;x++) {
    for(y=0;y<rows;y++) {
      cind=y+x*rows;
      cind2=x*fields+y*cols*fields+field;
      
      /* emark flags for edges */
      /*
       *       5 4 6
       *       1 0 2
       *       9 8 A
       */
      emark=0;
      if(y==0)      emark+=1;
      if(y==rows-1) emark+=2;
      if(x==0)      emark+=4;
      if(x==cols-1) emark+=8;
      
      /* Determine which pixel we're at right now */
      tcomp=((y+yoff)%2)+2*((x+xoff)%2);
      switch(tcomp) {
      case 0:
	/* At the RED pixel */
	gs=bs=0;
	rgb[cind+red]=field_ptr[cind2]/255.0;      /* Red */
	if(emark==0) {
	  gs+=field_ptr[cind2-bayer_d];
	  gs+=field_ptr[cind2+bayer_d];
	  gs+=field_ptr[cind2-bayer_r];
	  gs+=field_ptr[cind2+bayer_r];
	  rgb[cind+green]=gs/255.0/4.0;           /* Green */
	  bs+=field_ptr[cind2-bayer_rd];
	  bs+=field_ptr[cind2-bayer_ld];
	  bs+=field_ptr[cind2+bayer_ld];
	  bs+=field_ptr[cind2+bayer_rd];
	  rgb[cind+blue]=bs/255.0/4.0;            /* Blue */
	} else {
	  ng=nb=0;
	  if(!(emark&1))  {gs+=field_ptr[cind2-bayer_d];ng++;}
	  if(!(emark&2))  {gs+=field_ptr[cind2+bayer_d];ng++;}
	  if(!(emark&4))  {gs+=field_ptr[cind2-bayer_r];ng++;}
	  if(!(emark&8))  {gs+=field_ptr[cind2+bayer_r];ng++;}
	  rgb[cind+green]=gs/255.0/ng;              /* Green */
	  if(!(emark&5))  {bs+=field_ptr[cind2-bayer_rd];nb++;}
	  if(!(emark&9))  {bs+=field_ptr[cind2-bayer_ld];nb++;}
	  if(!(emark&6))  {bs+=field_ptr[cind+bayer_ld];nb++;}
	  if(!(emark&10)) {bs+=field_ptr[cind+bayer_rd];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;               /* Blue */
	}
	break;
      case 1:
	/* At the GREEN1 pixel */
	rs=bs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind2-bayer_d];
	  rs+=field_ptr[cind2+bayer_d];
	  rgb[cind+red]=rs/255.0/2.0;               /* Red */
	  gs+=field_ptr[cind2-bayer_rd];
	  gs+=field_ptr[cind2+bayer_ld];
	  gs+=4.0*field_ptr[cind2];
	  gs+=field_ptr[cind2-bayer_ld];
	  gs+=field_ptr[cind2+bayer_rd];
	  rgb[cind+green]=gs/255.0/8.0;             /* Green */
	  bs+=field_ptr[cind2-bayer_r];
	  bs+=field_ptr[cind2+bayer_r];
	  rgb[cind+blue]=bs/255.0/2.0;              /* Blue */
	} else {
	  nr=nb=ng=0;
	  if(!(emark&1))  {rs+=field_ptr[cind2-bayer_d];nr++;}
	  if(!(emark&2))  {rs+=field_ptr[cind2+bayer_d];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&5))  {gs+=field_ptr[cind2-bayer_rd];ng++;}
	  if(!(emark&6))  {gs+=field_ptr[cind2+bayer_ld];ng++;}
	  gs+=4.0*field_ptr[cind2];ng+=4;
	  if(!(emark&9))  {gs+=field_ptr[cind2-bayer_ld];ng++;}
	  if(!(emark&10)) {gs+=field_ptr[cind2+bayer_rd];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  if(!(emark&4))  {bs+=field_ptr[cind2-bayer_r];nb++;}
	  if(!(emark&8))  {bs+=field_ptr[cind2+bayer_r];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;              /* Blue */
	}
	break;
      case 2:
	/* At the GREEN2 pixel */
	rs=bs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind2-bayer_r];
	  rs+=field_ptr[cind2+bayer_r];
	  rgb[cind+red]=rs/255.0/2.0;               /* Red */
	  gs+=field_ptr[cind2-bayer_rd];
	  gs+=field_ptr[cind2+bayer_ld];
	  gs+=4.0*field_ptr[cind2];
	  gs+=field_ptr[cind2-bayer_ld];
	  gs+=field_ptr[cind2+bayer_rd];
	  rgb[cind+green]=gs/255.0/8.0;             /* Green */
	  bs+=field_ptr[cind2-bayer_d];
	  bs+=field_ptr[cind2+bayer_d];
	  rgb[cind+blue]=bs/255.0/2.0;              /* Blue */
	} else {
	  nr=nb=ng=0;
	  if(!(emark&4))  {rs+=field_ptr[cind2-bayer_r];nr++;}
	  if(!(emark&8))  {rs+=field_ptr[cind2+bayer_r];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&5))  {gs+=field_ptr[cind2-bayer_rd];ng++;}
	  if(!(emark&6))  {gs+=field_ptr[cind2+bayer_ld];ng++;}
	  gs+=4.0*field_ptr[cind2];ng+=4;
	  if(!(emark&9))  {gs+=field_ptr[cind2-bayer_ld];ng++;}
	  if(!(emark&10)) {gs+=field_ptr[cind2+bayer_rd];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  if(!(emark&1))  {bs+=field_ptr[cind2-bayer_d];nb++;}
	  if(!(emark&2))  {bs+=field_ptr[cind2+bayer_d];nb++;}
	  rgb[cind+blue]=bs/255.0/nb;              /* Blue */
	}
	  break;
      case 3:
	/* At the BLUE pixel */
	rs=gs=0;
	if(emark==0) {
	  rs+=field_ptr[cind2-bayer_rd];
	  rs+=field_ptr[cind2-bayer_ld];
	  rs+=field_ptr[cind2+bayer_ld];
	  rs+=field_ptr[cind2+bayer_rd];
	  rgb[cind+red]=rs/255.0/4.0;           /* Red */
	  gs+=field_ptr[cind2-bayer_d];
	  gs+=field_ptr[cind2+bayer_d];
	  gs+=field_ptr[cind2-bayer_r];
	  gs+=field_ptr[cind2+bayer_r];
	  rgb[cind+green]=gs/255.0/4.0;         /* Green */
	  rgb[cind+blue]=field_ptr[cind2]/255.0; /* Blue */
	} else {
	  nr=ng=0;
	  if(!(emark&5))  {rs+=field_ptr[cind2-bayer_rd];nr++;}
	  if(!(emark&9))  {rs+=field_ptr[cind2-bayer_ld];nr++;}
	  if(!(emark&6))  {rs+=field_ptr[cind2+bayer_ld];nr++;}
	  if(!(emark&10)) {rs+=field_ptr[cind2+bayer_rd];nr++;}
	  rgb[cind+red]=rs/255.0/nr;               /* Red */
	  if(!(emark&1))  {gs+=field_ptr[cind2-bayer_d];ng++;}
	  if(!(emark&2))  {gs+=field_ptr[cind2+bayer_d];ng++;}
	  if(!(emark&4))  {gs+=field_ptr[cind2-bayer_r];ng++;}
	  if(!(emark&8))  {gs+=field_ptr[cind2+bayer_r];ng++;}
	  rgb[cind+green]=gs/255.0/ng;             /* Green */
	  rgb[cind+blue]=field_ptr[cind2]/255.0;    /* Blue */
	}
	break;
      }
    }
  }
}
/*
 * Convert a raw Bayer buffer to an RGB buffer
 * by subsampling.
 *
 *  field_ptr pointer to image buffer (assumed to be
 *            double the size of bf_rgb)
 *  bf_rgb    new rgb buffer 
 *  bpattern  Bayer pattern type
 *
 */

void bayer_to_rgb_ds(unsigned char *field_ptr,buffer *bf_rgb,int bpattern) {
  int rows,cols,x,y;
  int cind,cind2,red,green,blue;
  fpnum *rgb;

  rows=bf_rgb->rows;
  cols=bf_rgb->cols;
  /* Get pointer to RGB buffer  */
  rgb=bf_rgb->data;
  /* Offsets */
  red=0;
  green=rows*cols;
  blue=2*rows*cols;

  switch(bpattern) {
  case BAYER_BGGR:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=2*y+4*x*rows;
 	rgb[cind+red]   = field_ptr[cind2+1+2*rows]/255.0;
 	rgb[cind+green] = (field_ptr[cind2+1]+
 			   field_ptr[cind2+2*rows])/510.0;
 	rgb[cind+blue]  = field_ptr[cind2]/255.0;
      }
    }
    break;
  case BAYER_GRBG:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=2*y+4*x*rows;
	rgb[cind+red]   = field_ptr[cind2+2*rows]/255.0;
	rgb[cind+green] = (field_ptr[cind2]+
			   field_ptr[cind2+1+2*rows])/510.0;
	rgb[cind+blue]  = field_ptr[cind2+1]/255.0;
      }
    }
    break;
  case BAYER_RGGB:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=2*y+4*x*rows;
	rgb[cind+red]   = field_ptr[cind2]/255.0;
	rgb[cind+green] = (field_ptr[cind2+1]+
			   field_ptr[cind2+2*rows])/510.0;
	rgb[cind+blue]  = field_ptr[cind2+1+2*rows]/255.0;
      }
    }
    break;
  case BAYER_GBRG:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {      
	cind=y+x*rows;
	cind2=2*y+4*x*rows;
 	rgb[cind+red]   = field_ptr[cind2+1]/255.0;
 	rgb[cind+green] = (field_ptr[cind2]+
 			   field_ptr[cind2+1+2*rows])/510.0;
 	rgb[cind+blue]  = field_ptr[cind2+2*rows]/255.0;
      }
    }
    break;
  }
}
/*
 * Convert one field of a raw Bayer stereo buffer to an RGB buffer
 * by subsampling. This function assumes sensor columns
 * alternatingly come from left and right fields.
 * This function also transposes input and output images.
 * 
 *  field_ptr    pointer to image buffer (assumed to be
 *               double the size of bf_rgb)
 *  bf_rgb       new rgb buffer
 *  bpattern     Bayer pattern type
 *  field        0 for left, 1 for right field
 */
void bayer_to_rgb_stereo16_ds(unsigned char *field_ptr,buffer *bf_rgb,int bpattern,int field) {
  int rows,cols,x,y;
  int cind,cind2,red,green,blue;
  fpnum *rgb;

  rows=bf_rgb->rows;
  cols=bf_rgb->cols;
  /* Get pointer to RGB buffer  */
  rgb=bf_rgb->data;
  /* Offsets */
  red=0;
  green=rows*cols;
  blue=2*rows*cols;

  switch(bpattern) {
  case BAYER_BGGR:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=4*x+8*y*cols+field;
 	rgb[cind+red]   = field_ptr[cind2+2+4*cols]/255.0;
 	rgb[cind+green] = (field_ptr[cind2+2]+
 			   field_ptr[cind2+4*cols])/510.0;
 	rgb[cind+blue]  = field_ptr[cind2]/255.0;
      }
    }
    break;
  case BAYER_GRBG:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=4*x+8*y*cols+field;
	rgb[cind+red]   = field_ptr[cind2+4*cols]/255.0;
	rgb[cind+green] = (field_ptr[cind2]+
			   field_ptr[cind2+2+4*cols])/510.0;
	rgb[cind+blue]  = field_ptr[cind2+2]/255.0;
      }
    }
    break;
  case BAYER_RGGB:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=4*x+8*y*cols+field;
	rgb[cind+red]   = field_ptr[cind2]/255.0;
	rgb[cind+green] = (field_ptr[cind2+2]+
			   field_ptr[cind2+4*cols])/510.0;
	rgb[cind+blue]  = field_ptr[cind2+2+4*cols]/255.0;
      }
    }
    break;
  case BAYER_GBRG:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {      
	cind=y+x*rows;
	cind2=4*x+8*y*cols+field;
 	rgb[cind+red]   = field_ptr[cind2+2]/255.0;
 	rgb[cind+green] = (field_ptr[cind2]+
 			   field_ptr[cind2+2+4*cols])/510.0;
 	rgb[cind+blue]  = field_ptr[cind2+4*cols]/255.0;
      }
    }
    break;
  }
}
/*
 * Convert a field of a possibly multi-field raw Bayer buffer to an
 * RGB buffer by interpolation with an [1 2 1]/2 kernel along
 * each dimension. Data is also scaled by 1.0/255.0
 * This function also transposes input and output images.
 *
 *  field_ptr pointer to image buffer (assumed to be
 *            the same size as bf_rgb)
 *  bf_rgb    pointer to new rgb buffer
 *  bpattern  Bayer pattern type
 *  fields    Number of fields
 *  field     Field to extract, 0/1 for left/right etc
 *
 */
void bayer_to_rgb_fields_ds(unsigned char *field_ptr,buffer *bf_rgb,int bpattern,int fields,int field) {
  int rows,cols,x,y;
  int cind,cind2,red,green,blue;
  int cell1,cell2,cell3;
  fpnum *rgb;

  rows=bf_rgb->rows;
  cols=bf_rgb->cols;
  /* Get pointer to RGB buffer  */
  rgb=bf_rgb->data;
  /* Offsets */
  red=0;
  green=rows*cols;
  blue=2*rows*cols;
  /* Bayer cell layout:
   *     0 1
   *     2 3
   */
  cell1=fields;
  cell2=2*fields*cols;
  cell3=fields+2*fields*cols;

  switch(bpattern) {
  case BAYER_BGGR:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=2*fields*x+4*fields*y*cols+field;
 	rgb[cind+red]   = field_ptr[cind2+cell3]/255.0;
 	rgb[cind+green] = (field_ptr[cind2+cell1]+
 			   field_ptr[cind2+cell2])/510.0;
 	rgb[cind+blue]  = field_ptr[cind2]/255.0;
      }
    }
    break;
  case BAYER_GRBG:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=2*fields*x+4*fields*y*cols+field;
	rgb[cind+red]   = field_ptr[cind2+cell2]/255.0;
	rgb[cind+green] = (field_ptr[cind2]+
			   field_ptr[cind2+cell3])/510.0;
	rgb[cind+blue]  = field_ptr[cind2+cell1]/255.0;
      }
    }
    break;
  case BAYER_RGGB:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	cind=y+x*rows;
	cind2=2*fields*x+4*fields*y*cols+field;
	rgb[cind+red]   = field_ptr[cind2]/255.0;
	rgb[cind+green] = (field_ptr[cind2+cell1]+
			   field_ptr[cind2+cell2])/510.0;
	rgb[cind+blue]  = field_ptr[cind2+cell3]/255.0;
      }
    }
    break;
  case BAYER_GBRG:
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {      
	cind=y+x*rows;
	cind2=2*fields*x+4*fields*y*cols+field;
 	rgb[cind+red]   = field_ptr[cind2+cell1]/255.0;
 	rgb[cind+green] = (field_ptr[cind2]+
 			   field_ptr[cind2+cell3])/510.0;
 	rgb[cind+blue]  = field_ptr[cind2+cell2]/255.0;
      }
    }
    break;
  }
}
