/*
**  dcam.c
**
**  Routines for converting a DCAM buffer to a Matlab RGB image
*/
#include <stdlib.h>
#include "dcam.h"

/*
 *  Convert RGB buffer to matlab
 *  (just flip dimensions and convert to double)
 */ 
void dcam_rgb_to_matlab(unsigned char *field_ptr,buffer *bf_image) {
int rows,cols,ndim;
int x,y,m;
fpnum *image;

 rows  = bf_image->rows;
 cols  = bf_image->cols;
 ndim  = bf_image->ndim;
 image = bf_image->data;

  for(m=0;m<ndim;m++) {
    for(x=0;x<cols;x++) {
      for(y=0;y<rows;y++) {
	image[y+x*rows+m*rows*cols]=(double)field_ptr[m+x*ndim+y*ndim*cols]/255.0;
      }
    }
  }
}
/*
 *  Convert YUV444 buffer to matlab
 *  Also flips dimensions and converts to double.
 */ 
void dcam_yuv444_to_matlab(unsigned char *field_ptr,buffer *bf_image) {
int rows,cols,ndim;
int x,y,m;
double M[]={1.00563080024661, 0.00928065000651, 1.40537989915084,
	    0.99680141120372, -0.33793486124619, -0.71516348375848,
	    1.00179933923475, 1.78724884666638, 0.00302259702047};

fpnum field_y,field_u,field_v,val;
int red,green,blue,field_ind;
fpnum *image;

 rows  = bf_image->rows;
 cols  = bf_image->cols;
 ndim  = bf_image->ndim;
 image = bf_image->data;

 /* Colour band offsets */
 red=0;
 green=rows*cols;
 blue=rows*cols*2;

 /* Pre-scale weights */
 for(m=0;m<9;m++) M[m]=M[m]/255.0;

 /* Data lies column-wise like this: U Y V U Y V... */
 for(x=0;x<cols;x++) {
   for(y=0;y<rows;y++) {
     field_ind=x*3+y*cols*3;
     field_u =(double)field_ptr[field_ind]-128.0;
     field_y =(double)field_ptr[field_ind+1];
     field_v =(double)field_ptr[field_ind+2]-128.0;
     /* Red colour */
     val=M[0]*field_y+M[1]*field_u+M[2]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+red  ]=val;

     /* Green colour */
     val=M[3]*field_y+M[4]*field_u+M[5]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+green]=val;

     /* Blue colour */
     val=M[6]*field_y+M[7]*field_u+M[8]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+blue ]=val;
   }
 }
}
/*
 *  Convert YUV422 buffer to matlab
 *  Also flips dimensions and converts to double.
 */ 
void dcam_yuv422_to_matlab(unsigned char *field_ptr,buffer *bf_image) {
int rows,cols,ndim;
int x,y,m;
double M[]={1.00563080024661, 0.00928065000651, 1.40537989915084,
	    0.99680141120372, -0.33793486124619, -0.71516348375848,
	    1.00179933923475, 1.78724884666638, 0.00302259702047};

fpnum field_y,field_u,field_v,val;
int red,green,blue,field_ind;
fpnum *image;

 rows  = bf_image->rows;
 cols  = bf_image->cols;
 ndim  = bf_image->ndim;
 image = bf_image->data;

 /* Colour band offsets */
 red=0;
 green=rows*cols;
 blue=rows*cols*2;

 /* Pre-scale weights */
 for(m=0;m<9;m++) M[m]=M[m]/255.0;

 /* Data lies column-wise like this: U Y V Y... */
 for(x=0;x<cols;x++) {
   for(y=0;y<rows;y++) {
     field_ind=x*2+y*cols*2;
     if(x%2) {
       /* At odd pixel */
       field_u=(double)field_ptr[field_ind-2];
       if(x<cols-1) {
	 field_u=(field_u+(double)field_ptr[field_ind+2])/2.0;
       }
       field_v=(double)field_ptr[field_ind];
     } else {
       /* At even pixel */
       field_u=(double)field_ptr[field_ind];

       field_v=(double)field_ptr[field_ind+2];
       if(x>0) { field_v=(field_v+(double)field_ptr[field_ind-2])/2.0;}
     }

     field_y =(double)field_ptr[field_ind+1];

     /* Remove offset */
     field_u=field_u-128.0;
     field_v=field_v-128.0;
     
     /* Red colour */
     val=M[0]*field_y+M[1]*field_u+M[2]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+red  ]=val;

     /* Green colour */
     val=M[3]*field_y+M[4]*field_u+M[5]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+green]=val;

     /* Blue colour */
     val=M[6]*field_y+M[7]*field_u+M[8]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+blue ]=val;
    }
  }
}
/*
 *  Convert YUV411 buffer to matlab
 *  Also flips dimensions and converts to double.
 */ 
void dcam_yuv411_to_matlab(unsigned char *field_ptr,buffer *bf_image) {
int rows,cols,ndim;
int x,y,m;
double M[]={1.00563080024661, 0.00928065000651, 1.40537989915084,
	    0.99680141120372, -0.33793486124619, -0.71516348375848,
	    1.00179933923475, 1.78724884666638, 0.00302259702047};

fpnum field_y,field_u=0,field_v=0,wsum,val;
int red,green,blue,field_ind;
fpnum *image;

 rows  = bf_image->rows;
 cols  = bf_image->cols;
 ndim  = bf_image->ndim;
 image = bf_image->data;

 /* Colour band offsets */
 red=0;
 green=rows*cols;
 blue=rows*cols*2;

 /* Pre-scale weights */
 for(m=0;m<9;m++) M[m]=M[m]/255.0;

 /* Data lies column-wise like this: U Y Y V Y Y U Y Y V Y Y... */
 /* Output:                            0 1   2 3   4 5   6 7... */
 for(x=0;x<cols;x++) {
   for(y=0;y<rows;y++) {
     field_ind=x*3/2+y*cols*3/2+1; /* Index of Y pixel */
     field_y=(double)field_ptr[field_ind];
     switch(x%4) {
     case 0:
       field_u=(double)field_ptr[field_ind-1];wsum=1;
       if(x<cols-4) { field_u+=(double)field_ptr[field_ind+5];wsum++;}
       field_u=field_u/wsum-128.0;

       field_v=0;wsum=0;
       if(x>3) { field_v+=(double)field_ptr[field_ind-4];wsum++;}
       if(x<cols-1) { field_v+=(double)field_ptr[field_ind+2];wsum++;}
       field_v=field_v/wsum-128.0;
       break;
     case 1:
       field_u=(double)field_ptr[field_ind-2];wsum=1;
       if(x<cols-4) {field_u +=(double)field_ptr[field_ind+4];wsum++;}
       field_u=field_u/wsum-128.0;

       field_v=0;wsum=0;
       if(x>3) {field_v+=(double)field_ptr[field_ind-5];wsum++;}
       if(x<cols-1) {field_v+=(double)field_ptr[field_ind+1];wsum++;}
       field_v=field_v/wsum-128.0;
       break;
     case 2:
       field_u=(double)field_ptr[field_ind-4];wsum=1;
       if(x<cols-3) { field_u+=(double)field_ptr[field_ind+2];wsum++;}
       field_u=field_u/wsum-128.0;

       field_v=(double)field_ptr[field_ind-1];wsum=1;
       if(x<cols-4) {field_v+=(double)field_ptr[field_ind+5];wsum++;}
       field_v=field_v/wsum-128.0;
       break;
     case 3:
       field_u=(double)field_ptr[field_ind-5];wsum=1;
       if(x<cols-1) {field_u+=(double)field_ptr[field_ind+1];wsum++;}
       field_u=field_u/wsum-128.0;
       
       field_v=(double)field_ptr[field_ind-2];wsum=1;
       if(x<cols-4) {field_v+=(double)field_ptr[field_ind+4];wsum++;}
       field_v=field_v/wsum-128.0;
     }
      /* Red colour */
     val=M[0]*field_y+M[1]*field_u+M[2]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+red  ]=val;

     /* Green colour */
     val=M[3]*field_y+M[4]*field_u+M[5]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+green]=val;

     /* Blue colour */
     val=M[6]*field_y+M[7]*field_u+M[8]*field_v;
     if(val<0) val=0;if(val>1) val=1;
     image[y+x*rows+blue ]=val;
   }
 }
}
