/*
Copyright (C) 2006 Pedro Felzenszwalb

Modified June 05 2007 by Matthew Baumann for use in the
HBackground homogeneous background detector.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/

#ifndef SEGMENT_IMAGE
#define SEGMENT_IMAGE

#include <cstdlib>

//#include <vector>

#include <image.h>
#include <misc.h>
#include <filter.h>
#include "segment-graph.h"

namespace segment{

//const float PIXEL_THRESHOLD = 0.3;
//const float BORDER_THRESHOLD = 0.75;

// random color
rgb random_rgb(){ 
  rgb c;
  double r;
  
  c.r = (uchar)random();
  c.g = (uchar)random();
  c.b = (uchar)random();

  return c;
}

// dissimilarity measure between pixels
static inline float diff(image<float> *r, image<float> *g, image<float> *b,
			 int x1, int y1, int x2, int y2) {
  return sqrt(square(imRef(r, x1, y1)-imRef(r, x2, y2)) +
	      square(imRef(g, x1, y1)-imRef(g, x2, y2)) +
	      square(imRef(b, x1, y1)-imRef(b, x2, y2)));
}

/*
 * Segment an image
 *
 * Returns a color image representing the segmentation.
 *
 * im: image to segment.
 * sigma: to smooth the image.
 * c: constant for treshold function.
 * min_size: minimum component size (enforced by post-processing stage).
 * num_ccs: number of connected components in the segmentation.
 */
image<rgb> *segment_image(image<rgb> *im, float sigma, float c, int min_size,
			  int *num_ccs) {
  int width = im->width();
  int height = im->height();

  image<float> *r = new image<float>(width, height);
  image<float> *g = new image<float>(width, height);
  image<float> *b = new image<float>(width, height);

  // smooth each color channel  
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      imRef(r, x, y) = imRef(im, x, y).r;
      imRef(g, x, y) = imRef(im, x, y).g;
      imRef(b, x, y) = imRef(im, x, y).b;
    }
  }
  image<float> *smooth_r = smooth(r, sigma);
  image<float> *smooth_g = smooth(g, sigma);
  image<float> *smooth_b = smooth(b, sigma);
  delete r;
  delete g;
  delete b;
 
  // build graph
  edge *edges = new edge[width*height*4];
  int num = 0;
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      if (x < width-1) {
	edges[num].a = y * width + x;
	edges[num].b = y * width + (x+1);
	edges[num].w = diff(smooth_r, smooth_g, smooth_b, x, y, x+1, y);
	num++;
      }

      if (y < height-1) {
	edges[num].a = y * width + x;
	edges[num].b = (y+1) * width + x;
	edges[num].w = diff(smooth_r, smooth_g, smooth_b, x, y, x, y+1);
	num++;
      }

      if ((x < width-1) && (y < height-1)) {
	edges[num].a = y * width + x;
	edges[num].b = (y+1) * width + (x+1);
	edges[num].w = diff(smooth_r, smooth_g, smooth_b, x, y, x+1, y+1);
	num++;
      }

      if ((x < width-1) && (y > 0)) {
	edges[num].a = y * width + x;
	edges[num].b = (y-1) * width + (x+1);
	edges[num].w = diff(smooth_r, smooth_g, smooth_b, x, y, x+1, y-1);
	num++;
      }
    }
  }
  delete smooth_r;
  delete smooth_g;
  delete smooth_b;

  // segment
  universe *u = segment_graph(width*height, num, edges, c);
  
  // post process small components
  for (int i = 0; i < num; i++) {
    int a = u->find(edges[i].a);
    int b = u->find(edges[i].b);
    if ((a != b) && ((u->size(a) < min_size) || (u->size(b) < min_size)))
      u->join(a, b);
  }
  delete [] edges;
  *num_ccs = u->num_sets();

  image<rgb> *output = new image<rgb>(width, height);

  // pick random colors for each component
  rgb *colors = new rgb[width*height];
  for (int i = 0; i < width*height; i++)
    colors[i] = random_rgb();
  
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      int comp = u->find(y * width + x);
      imRef(output, x, y) = colors[comp];
    }
  }  

  delete [] colors;  
  delete u;

  return output;
}

/*
 * Segment an image, and check for an homogeneous background
 *
 * Returns a boolean: true for homogeneous background, false otherwise
 *
 * im: image to segment.
 * sigma: to smooth the image.
 * c: constant for treshold function.
 * min_size: minimum component size (enforced by post-processing stage).
 * num_ccs: number of connected components in the segmentation.
 * pixel_threshold: min proportion of total image pixesl that a segment
 *     must occupy to be an homogeneous background
 * border_threshold: min proportion of border pixels that a segment must
 *     occupy to be an homogeneous background.
 * output: a reference to an image into which the segmented result image 
 *     is placed.
 *
 */
bool hbackground(image<rgb> *im, float sigma, float c, int min_size, int *num_ccs, double pixel_threshold, double border_threshold, image<rgb>* &output) 
{
   bool result = false;


  int width = im->width();
  int height = im->height();

  image<float> *r = new image<float>(width, height);
  image<float> *g = new image<float>(width, height);
  image<float> *b = new image<float>(width, height);

  // smooth each color channel  
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      imRef(r, x, y) = imRef(im, x, y).r;
      imRef(g, x, y) = imRef(im, x, y).g;
      imRef(b, x, y) = imRef(im, x, y).b;
    }
  }
  image<float> *smooth_r = smooth(r, sigma);
  image<float> *smooth_g = smooth(g, sigma);
  image<float> *smooth_b = smooth(b, sigma);
  delete r;
  delete g;
  delete b;
 
  // build graph
  edge *edges = new edge[width*height*4];
  int num = 0;
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      if (x < width-1) {
	edges[num].a = y * width + x;
	edges[num].b = y * width + (x+1);
	edges[num].w = diff(smooth_r, smooth_g, smooth_b, x, y, x+1, y);
	num++;
      }

      if (y < height-1) {
	edges[num].a = y * width + x;
	edges[num].b = (y+1) * width + x;
	edges[num].w = diff(smooth_r, smooth_g, smooth_b, x, y, x, y+1);
	num++;
      }

      if ((x < width-1) && (y < height-1)) {
	edges[num].a = y * width + x;
	edges[num].b = (y+1) * width + (x+1);
	edges[num].w = diff(smooth_r, smooth_g, smooth_b, x, y, x+1, y+1);
	num++;
      }

      if ((x < width-1) && (y > 0)) {
	edges[num].a = y * width + x;
	edges[num].b = (y-1) * width + (x+1);
	edges[num].w = diff(smooth_r, smooth_g, smooth_b, x, y, x+1, y-1);
	num++;
      }
    }
  }
  delete smooth_r;
  delete smooth_g;
  delete smooth_b;

  // segment
  universe *u = segment_graph(width*height, num, edges, c);
  
  // post process small components
  for (int i = 0; i < num; i++) {
    int a = u->find(edges[i].a);
    int b = u->find(edges[i].b);
    if ((a != b) && ((u->size(a) < min_size) || (u->size(b) < min_size)))
      u->join(a, b);
  }
  delete [] edges;
  *num_ccs = u->num_sets();

  int num_components = width*height;

  //count number of pixels in each component
  int* component_counts = new int[num_components];
  
  //count number of border pixels in each component
  int* component_border_counts = new int[num_components];
  
  for(int i = 0; i < num_components; i++)
  {
  	component_counts[i] = 0;
  	component_border_counts[i] = 0;
  }
  
  //iterate across the image and compute the above counts
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      int comp = u->find(y * width + x);
      component_counts[comp]++;
      //printf("comp = %d\n",comp);
      
      if(y == 0 || y == height-1 || x == 0 || x == width-1)
      {
      	component_border_counts[comp]++;
      	//printf("Edge [%d,%d]\n",x,y);
      }
    }
  }
  
  //std::vector<int> segment_counts;
  //std::vector<int> segment_border_counts;
  //std::vector<float> area_ratio;
  //std::vector<float> border_ratio;
 
 
  int border_total = 2*width + 2 * height - 4;  //total number of pixels of the border
  int total = width*height;						//total number of pixels
  
  for(int j = 0; j < num_components; j++)
  {
  	if(component_counts[j] > 0)
  	{
  		//segment_counts.push_back(component_counts[j]);
  		//segment_border_counts.push_back(component_border_counts[j]);
  		
  		//area_ratio.push_back( (float) component_counts[j] / (float) total);  //what ratio of the tota imag pixels does this region possess?
  		//border_ratio.push_back( (float) component_border_counts / (float) border_total );  //what ratio of the border does it occupy?
  		
  		float ratio = (float) component_counts[j] / (float) total;
  		float b_ratio = (float) component_border_counts[j] / (float) border_total;
  		
  		//printf("component %7d has count %7d, border count %5d, ratio %.3f, border ratio %.3f\n",j,component_counts[j],component_border_counts[j],ratio,b_ratio);
  		
  		if(ratio >= pixel_threshold && b_ratio >= border_threshold	)
  		{
  			result = true;
  			printf("Homogeneous Background found.\n");
  			printf("component %7d has count %7d, border count %5d, ratio %.3f, border ratio %.3f\n",j,component_counts[j],component_border_counts[j],ratio,b_ratio);
  			break;
  		}
  		else
  		{
  			//printf("Homogeneous Background not found.\n");
  			//printf("component %7d has count %7d, border count %5d, ratio %.3f, border ratio %.3f\n",j,component_counts[j],component_border_counts[j],ratio,b_ratio);
  		}
		//printf("Homogeneous Background not found\n");
  	}
  }

  if(!result)
  {
	printf("Homogeneous Background not found\n");
  }

	//
  //image<rgb> *output = new image<rgb>(width, height);
  output = new image<rgb>(width, height);
  
  if(output == NULL)
  {
  	printf("output isNULL in segment");
  }

  // pick random colors for each component
  rgb *colors = new rgb[width*height];
  for (int i = 0; i < width*height; i++)
    colors[i] = random_rgb();
  
  for (int y = 0; y < height; y++) {
    for (int x = 0; x < width; x++) {
      int comp = u->find(y * width + x);
      imRef(output, x, y) = colors[comp];
    }
  }  

  delete [] colors; 
  
  delete [] component_counts;
  delete [] component_border_counts;
  
  delete u;

  return result;
}

}

#endif
